#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os, subprocess


rotation_symbolic_compact_to_idxrot = {
        "":"",          "ff":"FF",      "ll":"LL",      "uu":"UU",
        "f":"F",        "f'":"B",       "fuu":"UUB",    "fll":"UUF",
        "ffu":"ULL",    "u":"U",        "u'":"D",       "ffu'":"UFF",
        "fu":"UL",      "f'u":"UR",     "fu'":"FD",     "f'u'":"LB",
        "f'l'":"DB",    "uf":"UF",      "fl":"LD",      "lu":"UB",
        "l'":"R",       "ffl'":"UUL",   "l":"L",        "ffl":"UUR",
    }
    
def indices_compact_to_index(indices, size):
    indices = tuple(indices)
    for b in range(size * size * size):
        bindices = (b % size, b // size % size, b // size // size)
        if bindices == indices:
            return b
    raise ValueError('Invalid block indices:', indices)
    
def index_compact_to_indices(index, size):
    return index % size, index // size % size, index // size // size
    
def block_compact_to_index_rot(block, size):
    # every block is stored as 4-tuple (x,y,z,sym), where sym is a symbolic rotation
    block = block.strip().split(',')
    x, y, z, rot = block
    x, y, z = int(x), int(y), int(z)
    x, y, z = y, size-1 - z, size-1 - x
    indices = x, y, z
    try:
        rot = rotation_symbolic_compact_to_idxrot[rot]
    except KeyError:
        raise ValueError('Invalid block rotation:', rot)
    index = indices_compact_to_index(indices, size)
    return '{}-{}'.format(index, rot)
    
def blocks_compact_to_index_rot(blocks, size):
    def blocks_perm(index):
        x, y, z = index_compact_to_indices(index, size)
        x, y, z = size-1 - z, x, size-1 - y
        index = indices_compact_to_index((x, y, z), size)
        return blocks[index]
    blocks = blocks.strip().split(' ')
    for i, block in enumerate(blocks):
        blocks[i] = block_compact_to_index_rot(block, size)
    blocks = [blocks_perm(i) for i, unused_b in enumerate(blocks)]
    return ' '.join(blocks)
    
def parse_initial_state(state):
    msf, blocks = state.split(':', 1)
    model, size, bformat = msf.split(' ', 2)
    if model != 'Cube':
        print('Unknown model:', repr(model))
        print('  Fallback:', 'Cube', 3, 'solved')
        return 'Cube', 3, 'solved'
    try:
        size = int(size)
    except ValueError:
        print('Unknown size:', repr(size))
        print('  Fallback:', 'Cube', 3, 'solved')
        return 'Cube', 3, 'solved'
    if bformat == 'identity':
        blocks = 'solved'
    elif bformat == 'blocks_compact':
        blocks = 'idx-rot: ' + blocks_compact_to_index_rot(blocks, size)
    else:
        blocks = 'solved'
    return model, size, blocks
    
def gconf_0_5_to_settings_1_0(settings_file):
    path_to_face = {#'/apps/pybik': None,
                    #'/apps/pybik/colors:': None,
                    '/apps/pybik/colors/0:': 0,
                    '/apps/pybik/colors/1:': 1,
                    '/apps/pybik/colors/2:': 2,
                    '/apps/pybik/colors/3:': 3,
                    '/apps/pybik/colors/4:': 4,
                    '/apps/pybik/colors/5:': 5,
                }
    path = '/apps/pybik'
    try:
        process = subprocess.Popen(['gconftool', '--recursive-list', path], stdout=subprocess.PIPE, universal_newlines=True)
    except OSError:
        return
    lines = process.communicate()[0]
    settings_dir = os.path.dirname(settings_file)
    if settings_dir and not os.path.exists(settings_dir):
        os.makedirs(settings_dir)
    with open(settings_file, 'wt', encoding='utf-8') as settings:
        face = 0
        for line in lines.splitlines():
            line = line.strip()
            face = path_to_face.get(line, face)
            if '=' not in line:
                facetype = None
                pattern = None
                imagefile = None
                imagemode = None
                continue
            key, value = line.split('=', 1)
            key = key.strip()
            value = value.strip()
            # pylint: disable=C0321
            if   key == 'saved_pos':        print('game.position =', value, file=settings)
            elif key == 'selection_mode':
                value = 'quadrant' if value == '0' else 'simple'
                print('draw.selection =', repr(value), file=settings)
            elif key == 'saved_moves':      print('game.moves =', repr('native: '+value), file=settings)
            elif key == 'lighting':         print('draw.lighting =', value=='true', file=settings)
            elif key == 'background_color': print('theme.bgcolor =', repr(value), file=settings)
            elif key == 'saved_state':
                if value:
                    mtype, size, blocks = parse_initial_state(value)
                    print('game.type =', repr(mtype), file=settings)
                    print('game.size =', (size, size, size), file=settings)
                    print('game.blocks =', repr(blocks), file=settings)
            elif key == 'color':            print('theme.face.%s.color =' % face, repr(value), file=settings)
            elif key == 'facetype':
                facetype = value
                if facetype == '1' and pattern is not None:
                    print('theme.face.%s.image =' % face, repr(pattern), file=settings)
                if facetype == '2' and imagefile is not None:
                    print('theme.face.%s.image =' % face, repr(imagefile), file=settings)
                if facetype == '2' and imagemode is not None:
                    print('theme.face.%s.mode =' % face, repr(imagemode), file=settings)
            elif key == 'pattern':
                if   value == '0': pattern = 'BEAMED EIGHTH NOTES.png'
                elif value == '1': pattern = 'diagonal-lines.png'
                elif value == '2': pattern = 'WHITE SMILING FACE.png'
                elif value == '3': pattern = 'BLACK SMILING FACE.png'
                elif value == '4': pattern = 'squares.png'
                elif value == '5': pattern = 'WHITE SUN WITH RAYS.png'
                if facetype == '1':
                    print('theme.face.%s.image =' % face, repr(pattern), file=settings)
            elif key == 'imagefile':
                imagefile = value
                if facetype == '2':
                    print('theme.face.%s.image =' % face, repr(imagefile), file=settings)
            elif key == 'imagemode':
                imagemode = 'tiled' if value != '1' else 'mosaic'
                if facetype == '2':
                    print('theme.face.%s.mode =' % face, repr(imagemode), file=settings)
            
def main():
    if len(sys.argv) != 2:
        sys.exit(1)
    gconf_0_5_to_settings_1_0(sys.argv[1])
    
    
if __name__ == '__main__':
    main()





#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
from os.path import dirname, basename, join, isfile, isdir
import re


def convert_symbols(sources):
    yield '\nfrom libc.stddef cimport ptrdiff_t'
    yield 'from libc.stdint cimport int32_t, intptr_t, int8_t, uint8_t'
    
    for abspath, relpath, source in sources:
        yield '\n\n# from {}:\n'.format(abspath)
        yield "cdef extern from '{}':".format(relpath)
        for match in re.finditer(r'^\#define[ \t]+(\w+).*?$', source, re.MULTILINE | re.DOTALL):
            if match.group(1).startswith('GL_'):
                yield match.expand(r'    enum: \1')
                
    for abspath, relpath, source in sources:
        yield '\n\n# from {}:\n'.format(abspath)
        for match in re.finditer(r'^(typedef[\s\w]*);.*?$', source, re.MULTILINE | re.DOTALL):
            mg1 = match.group(1)
            mg1 = mg1.replace('khronos_float_t', 'float')
            mg1 = mg1.replace('khronos_', '')
            if not mg1.endswith(('_t', '64')) and not (mg1[-2:].isalpha() and mg1[-2:].isupper()):
                yield 'c'+mg1
            
    yield '\n\ncdef extern from *:'
    def const_types(types):
        for type in types.split():
            yield '    ctypedef {0}* const_{0}_ptr "const {0}*"'.format(type)
    yield from const_types('GLubyte GLboolean GLvoid GLchar GLfloat GLint GLshort '
                           'GLbyte GLuint GLushort GLclampf GLsizei GLenum void')
                           
    for abspath, relpath, source in sources:
        yield '\n\n# from {} with:\n'.format(abspath)
        yield "cdef extern from '{}':".format(relpath)
        for match in re.finditer(
                r'^(?:GLAPI|GL_APICALL)([\s\w*]*?)(?:GLAPIENTRY|APIENTRY|GL_APIENTRY)([^(]*)\(([^)]*)\);(.*?)$',
                source, re.MULTILINE | re.DOTALL):
            mg2s2 = match.group(2).strip()[-2:]
            if mg2s2.isalpha() and mg2s2.isupper():
                continue
            for mgf in (match.group(1).find, match.group(3).find):
                if not mgf('64') == mgf('GLsync') == mgf('GLDEBUGPROC') == -1:
                    break
            else:
                if match.group(3).strip() == 'void':
                    template = r'    cdef\1\2()\4'
                else:
                    template = r'    cdef\1\2(\3)\4'
                yield match.expand(template).replace('const ', '').replace('const*', '*') \
                                            .replace(' in,', ' in_,').replace('/*', '#/*')
                
def read_url(url, headerfilename):
    import urllib.request
    text = urllib.request.urlopen(url).read().decode('utf-8')
    return [(url, headerfilename, text)]
    
def read_headers(path, headerfiles):
    sources = []
    for headerfilename in headerfiles:
        absheaderfilename = join(path, headerfilename) if isdir(path) else path
        with open(absheaderfilename, 'rt', encoding='utf-8') as headerfile:
            sources.append((absheaderfilename, headerfilename, headerfile.read()))
    return sources
    
def create_pxd(sources, pxdfilename):
    with open(pxdfilename, 'wt', encoding='utf-8') as pxdfile:
        print('# {}'.format(pxdfilename), file=pxdfile)
        print('# generated with {}'.format(sys.argv[0]), file=pxdfile)
        for token in convert_symbols(sources):
            print(token, file=pxdfile)
            
def usage(status):
    print('usage:\n'
          '{0} [-h|--help]\n'
          '{0} gl|gles2|gles3\n'
          '{0} filename\n'
          '{0} url\n'
          .format(basename(sys.argv[0])))
    sys.exit(status)
    
def main():
    args = sys.argv[1:]
    if not args or args[0] in ['-h', '--help']:
        usage(0)
    if len(args) != 1:
        print('wrong number of arguments')
        usage(1)
    includedir = '/usr/include'
    if args[0] == 'gl':
        sources = read_headers(includedir, ['GL/gl.h', 'GL/glext.h'])
        create_pxd(sources, 'pybiklib/gl.pxd')
    elif args[0] == 'gles2':
        sources = read_headers(includedir, ['GLES2/gl2.h', 'GLES2/gl2ext.h'])
        create_pxd(sources, 'pybiklib/gl.pxd')
    elif args[0] == 'gles3':
        sources = read_headers(includedir, ['GLES3/gl3.h'])
        create_pxd(sources, 'pybiklib/gl.pxd')
    elif isfile(args[0]):
        sources = read_headers(args[0], ['GL/gl.h'])
        create_pxd(sources, 'pybiklib/gl.pxd')
    elif args[0].startswith('http://') or args[0].startswith('https://'):
        sources = read_url(args[0], 'GL/gl.h')
        create_pxd(sources, 'pybiklib/gl.pxd')
    else:
        print('wrong argument:', args[0])
        usage(1)
    

if __name__ == '__main__':
    main()
    




#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009, 2011-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.



import os, sys

N_ = lambda s: s

AUTHOR = 'B. Clausius'
CONTACT_EMAIL = 'barcc@gmx.de'
COPYRIGHT = 'Copyright © 2009-2013,  B. Clausius'
LICENSE_NAME = 'GPL-3+'
PACKAGE = 'pybik'
# Name of the application, probably should not be translated.
APPNAME = N_('Pybik')
VERSION = '1.1'
WEBSITE = 'https://launchpad.net/pybik/'
DOWNLOADS = 'https://launchpad.net/pybik/+download'
CONTACT_FILEBUG = 'https://bugs.launchpad.net/pybik/+filebug'


# The following two lines are replaced by setup.py
data_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data')
appdata_dir = data_dir

LOCALE_DIR   = os.path.join(data_dir, 'locale')
QT_LOCALE_DIR = os.path.join(sys.prefix, 'share', 'qt4', 'translations')
PIXMAP_DIR   = os.path.join(data_dir, 'pixmaps')
IMAGE_FILE   = os.path.join(PIXMAP_DIR, PACKAGE+'.png')
LICENSE_FILE = os.path.join(appdata_dir, 'GPL-3')
MODELS_DIR   = os.path.join(appdata_dir, 'models')
SCRIPT_DIR   = os.path.join(appdata_dir, 'plugins')
UI_DIR       = os.path.join(appdata_dir, 'ui')
TESTDATA_DIR = os.path.join(appdata_dir, 'tests')

def get_testdatafiles():
    return [f for f in os.listdir(TESTDATA_DIR) if not f.endswith('~')]
    
def get_config_home():
    return os.environ.get('XDG_CONFIG_HOME', '') or os.path.expanduser("~/.config")
def get_data_home():
    return os.environ.get('XDG_DATA_HOME', '') or os.path.expanduser("~/.local/share")
    
USER_CONFIG_DIR    = os.path.join(get_config_home(), PACKAGE)
USER_SETTINGS_FILE = os.path.join(USER_CONFIG_DIR, 'settings.conf')
USER_VERSION_FILE  = os.path.join(USER_CONFIG_DIR, 'version')
USER_SCRIPT_DIR    = os.path.join(get_data_home(), PACKAGE, 'plugins')

SHORT_DESCRIPTION = N_("3D Rubik's cube game")
LONG_DESCRIPTION = N_(
    'Pybik is an interactive, graphical, single player puzzle about the cube invented by'
    ' Ernő Rubik. Besides the cube the program can handle towers and bricks (non cubic'
    ' puzzles). Pybik also has solvers, pretty patterns and a collection of various'
    ' moves. The cube can be manipulated with the mouse or keyboard. You can change the'
    ' colors or images on the faces of the cube.'
)

LICENSE_INFO = N_(
            'This program is free software: you can redistribute it and/or modify '
            'it under the terms of the GNU General Public License as published by '
            'the Free Software Foundation, either version 3 of the License, or '
            '(at your option) any later version.\n\n'

            'This program is distributed in the hope that it will be useful, '
            'but WITHOUT ANY WARRANTY; without even the implied warranty of '
            'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the '
            'GNU General Public License for more details.')
LICENSE_FURTHER = N_(
# Text between "<" and ">" is expanded to a link by the program and should not be modified.
# Text between "<LINK|>" and "<|>" is the translatable text for the link.
            'Read the full text of the <text:FULL_LICENSE_TEXT|>GNU General Public License'
            '<|> or see <http://www.gnu.org/licenses/>.')
LICENSE_NOT_FOUND = N_(
            'You should have received a copy of the GNU General Public License'
            ' along with this program.  If not, see <http://www.gnu.org/licenses/>.')

def get_filebug_text():
    return _(
    # "Wishlist" is used on Launchpad for Importance, so this word should probably not be translated
    'If you find any bugs in Pybik or have a suggestion for an improvement then please '
    'submit a <{CONTACT_FILEBUG}|>bug report<|>. '
    'In the latter case you can mark the bug report as "Wishlist".'
    ).format(CONTACT_FILEBUG=CONTACT_FILEBUG)

TRANSLATION_TEXT = N_(
    'Translations are managed by the '
    '<https://translations.launchpad.net/+groups/launchpad-translators|>'
    'Launchpad translation group<|>.\n\n'
    'If you want help to translate Pybik to your language you can do it through '
    'the <https://translations.launchpad.net/pybik|>web interface<|>.\n\n'
    'Read more about <https://help.launchpad.net/Translations|>"Translating with Launchpad"<|> '
    'and <https://help.launchpad.net/Translations/StartingToTranslate|>"Starting to translate"<|>.')

def splitlines(description):
    line = None
    for word in description.split():
        if line is None:
            line = word
        elif len(line) + 1 + len(word) > 78:
            yield line
            line = word
        else:
            line = ' '.join((line, word))
    if line is not None:
        yield line
        
def wrap(text, sep='\n'):
    if type(text) in [list, tuple]:
        text, lines = text
    else:
        lines = None
    text = sep.join(splitlines(text))
    if lines:
        text += sep + sep + sep.join(lines)
    return text
    




#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009, 2011-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: textures.c
# Original copyright and license: 2003  John Darrington, GPL3+


import os
from glob import glob

# pylint: disable=W0614,W0401
from PyQt4.QtCore import Qt
from PyQt4.QtGui import *
# pylint: enable=W0614,W0401

from . import config

class Textures (object):
    max_size = 256
    stock_dir = os.path.join(config.UI_DIR, 'images')
    
    def __init__(self):
        self.stock_files = sorted(os.path.basename(f) for f in glob(os.path.join(self.stock_dir, '*')))
        self.stock_pixbuf = {'': self.create_dummy()}
        
    def get_stock_pixbuf(self, name):
        try:
            return self.stock_pixbuf[name]
        except KeyError:
            if name not in self.stock_files:
                return self.stock_pixbuf['']
            filename = os.path.join(self.stock_dir, name)
            self.stock_pixbuf[name] = pixbuf = self.create_pixbuf_from_file(filename)
            return pixbuf
            
    @classmethod
    def create_pixbuf_from_file(cls, filename):
        image = QImage(filename)
        # We must scale the image,  because Mesa/OpenGL insists on it being of size
        # 2^n ( where n is integer )
        width = image.width()
        height = image.height()
        scaled_width = cls.max_size
        while scaled_width > width:
            scaled_width //= 2
        scaled_height = cls.max_size
        while scaled_height > height:
            scaled_height //= 2
        return image.scaled(scaled_width, scaled_height, transformMode=Qt.SmoothTransformation)
        
    @classmethod
    def create_dummy(cls):
        image = QImage(1, 1, QImage.Format_ARGB32)
        image.fill(0)
        return image
        
textures = Textures()




#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009, 2011-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: main.c
# Original copyright and license: 1998, 2003--2004  John Darrington, GPL3+



import sys, os
from collections import namedtuple

from . import debug, config

N_ = lambda s: s


def format_opts(opts):
    opts = ((o + o.strip('--')[0].upper() if o.endswith('=') else o) for o in opts)
    return '  ' + ', '.join(opts)
    
def format_help(text, indent=0):
    try:
        width = int(os.environ['COLUMNS']) - 2
    except (KeyError, ValueError):
        width = 78
    width -= indent
    def split(text):
        lines = text.split('\n')
        for line in lines:
            res = None
            words = line.split(' ')
            for word in words:
                if res is None:
                    res = word
                elif len(res + word) + 1 <= width:
                    res = ' '.join((res, word))
                else:
                    yield res
                    res = word
            yield res
    return '\n'.ljust(indent+1).join(split(text))
    
def print_usage(unused_value, unused_opts):
    print('Usage:', sys.argv[0], '[options]')
    for args, unused_func, text in arg_info:
        if args is None:
            print()
            print(format_help(text))
        else:
            opts = format_opts(args)
            if len(opts) > 18:
                text = '\n' + text
            else:
                opts = opts.ljust(20)
            text = format_help(text, 20)
            print(opts + text)
    sys.exit(0)
    
def print_version(*unused_args):
    print(config.APPNAME, config.VERSION)
    print()
    print(config.COPYRIGHT.replace('©', '(C)')) #XXX: help2man cannot handle unicode
    print()
    print(config.wrap(config.LICENSE_INFO))
    print()
    print(config.wrap(config.LICENSE_NOT_FOUND))
    sys.exit(0)
    
def print_tests(*args):
    for filename in config.get_testdatafiles():
        print(filename)
    sys.exit(0)
    
def select_all_tests(value, opts):
    opts.setdefault('test', []).extend(config.get_testdatafiles())
    
debug_level_names = [__a[6:].lower() for __a in debug.__all__ if __a.startswith('DEBUG_')]
test_arg_names = ['write-y', 'write-yes', 'write-n', 'write-no', 'write-e', 'write-error',
                  'log-file', 'log-widgets']

arg_info = [
        (None, None, config.LONG_DESCRIPTION.replace('ő', 'o')), #XXX: help2man cannot handle unicode
        (None, None, 'Pybik can be fully controlled and configured via the graphical user interface.'
                     ' The options listed here are intended primarily for testing and debugging.'),
        (None, None, 'Options:'),
        (['-h', '--help'], print_usage,
                            'Show help message and exit'),
        (['--version'], print_version,
                            'Show version number and exit'),
        (['--config-file='], lambda value, opts: opts.update(config_file=value),
                            'Specify the configuration file to use instead of the default location'),
        (['--defaultconfig'], lambda value, opts: opts.update(defaultconfig=True),
                            'Print default settings to stdout and exit'),
        (['--shader='], lambda value, opts: opts.update(shader=value),
                            'Use a different shader'),
        (['--list-tests'], print_tests,
                            'List tests'),
        (['--test'], select_all_tests,
                            'Run all tests'),
        (['--test='], lambda value, opts: opts.setdefault('test', []).append(value),
                            'Run test T, can be used multiple times'),
        (['--test-args='], lambda value, opts: opts['test_args'].extend(a for a in value.split(',') if a),
                            'Comma-separated list of test arguments:\n{}'
                            .format(' '.join(test_arg_names))),
        (['--debug='], lambda value, opts: opts.setdefault('debug_flags', []).extend(
                                            f for f in value.split(',') if f),
                            'D is a comma-separated list of debug flags:\n{0}'
                            .format(' '.join(debug_level_names))),
        (None, None,  'Qt-Options, for a full list refer to the Qt-Reference,'
                ' but not all options make sense for this application:'),
        (['-style S'], None,
                            'sets the application GUI style. Possible values are'
                            ' motif, windows, and platinum.'),
        (['-widgetcount'], None,
                            'prints debug message at the end about number of widgets left'
                            ' undestroyed and maximum number of widgets existed at the same time'),
        (['-reverse'], None,
                            "sets the application's layout direction to left-to-right"),
        (['-graphicssystem G'], None,
                            'sets the backend to be used for on-screen widgets and pixmaps.'
                            ' Available options are raster and opengl.'),
    ]
    
def parse_args(args):
    arg_functs = {o: f for opts, f, h in arg_info for o in opts or [] if f is not None}
    opts = {
            'config_file': config.USER_SETTINGS_FILE,
            'defaultconfig': False,
            'shader': None,
            'pure_python': False,
            'test': [],
            'test_args': [],
            'ui_args': [],
        }
    for arg in args:
        try:
            index = arg.index('=')
        except ValueError:
            value = None
        else:
            arg, value = arg[:index+1], arg[index+1:]
        try:
            func = arg_functs[arg]
        except KeyError:
            opts['ui_args'].append(arg)
        else:
            func(value, opts)
            
    for a in opts['test_args']:
        if a not in test_arg_names:
            print('unknown test argument:', a)
            sys.exit(1)
        
    for d in opts.get('debug_flags', []):
        if d not in debug_level_names:
            print('unknown debug option:', d)
            sys.exit(1)
    opts.setdefault('debug_flags', None)
    
    class Opts: pass
    for k, v in opts.items():
        setattr(Opts, k, v)
    return Opts
    
def create_app(root_dir, args):
    import gettext
    from PyQt4.QtCore import QLocale, QTranslator, Qt
    from PyQt4.QtGui import QApplication, QPalette, QColor
    
    # initialize QApplication
    QApplication.setAttribute(Qt.AA_X11InitThreads)
    app = QApplication(args)
    args = app.arguments()[1:]
    if args:
        print('Unknown arguments:', ' '.join(args))
        sys.exit(1)
    app.setOrganizationName(config.PACKAGE)
    app.setApplicationName(config.APPNAME)
    app.setApplicationVersion(config.VERSION)
    
    # Workaround for whatsThis-Text (white text on light background)
    palette = app.palette()
    colorfg = palette.color(QPalette.Active, QPalette.ToolTipText)
    colorbg = palette.color(QPalette.Active, QPalette.ToolTipBase)
    valuefg = colorfg.value()
    valuebg = (valuefg + 255) // 2 if valuefg < 128 else valuefg // 2
    colorbg = QColor.fromHsv(colorbg.hue(), colorbg.saturation(), valuebg)
    palette.setColor(QPalette.Active, QPalette.ToolTipBase, colorbg)
    app.setPalette(palette)
    
    # initialize translation
    language = QLocale.system().name()
    # standard Qt translation, used for e.g. standard buttons and shortcut names
    translator = QTranslator()
    translator.load('qt_' + language, config.QT_LOCALE_DIR)
    app.installTranslator(translator)
    # the rest of the app use gettext for translation
    if root_dir == sys.prefix:
        # normal installation
        LOCALEDIR = None
    else:
        # different root, e.g. /usr/local, source directory
        LOCALEDIR = config.LOCALE_DIR
    t = gettext.translation(config.PACKAGE, LOCALEDIR, languages=[language], fallback=True)
    t.install(names=['ngettext'])
    return app, translator
    
def create_window(opts):
    # initialize settings
    from .settings import settings
    try:
        settings.load(opts.config_file)
    except EnvironmentError as e:
        error_message = N_('An error occurred while reading the settings:\n'
                        '{error_message}').format(error_message=e)
        debug.debug(error_message)
        settings.load('')
        from PyQt4.QtCore import QTimer
        QTimer.singleShot(0, lambda: window.error_dialog(_(error_message)))
        
    # create main window
    # The application module must be imported after a QApplication object is created
    from .application import MainWindow
    return MainWindow(opts)
    
    
def run(root_dir=None):
    opts = parse_args(sys.argv)
    
    if opts.debug_flags is not None:
        print(config.PACKAGE, config.VERSION)
        debug.set_flags(opts.debug_flags)
        print('debug flags:', *opts.debug_flags)
        debug.debug('Qt args:', opts.ui_args)
    
    try:
        import sip
        sip.setapi('QString', 2)
        sip.setapi('QVariant', 2)
        import PyQt4.QtCore
        import PyQt4.QtGui
        import PyQt4.QtOpenGL
    except ImportError as e:
        print('This program needs PyQt4:', e)
        sys.exit(1)
    
    if opts.defaultconfig:
        # settings needs Qt
        from .settings import settings
        settings.load('')
        settings.keystore.dump(sys.stdout, all=True)
        sys.exit(0)
        
    app, translator = create_app(root_dir, opts.ui_args)
    # keep the translator object for the lifetime of the app object, so that the translation works
    if opts.test:
        from pybiktest.testrunner import TestRunner
        for test in TestRunner.wrap(config.TESTDATA_DIR, opts.test, opts.test_args):
            opts.config_file = test.settings_file
            window = create_window(opts)
            test.run(window)
            app.exec()
    else:
        window = create_window(opts)
        app.exec()
    




#-*- coding:utf-8 -*-
# cython: profile=False
# distutils: language = c++

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: glarea-common.c
# Original copyright and license: 2003  John Darrington, GPL3+

# although this file is compiled with Python3 syntax, Cython needs at least division from __future__
from __future__ import print_function, division

# This line makes cython happy
global __name__, __package__    # pylint: disable=W0604
#px/__compiled = True
__compiled = False

#px/from libc.math cimport cos, sin, tan, atan2, M_PI
from math import cos, sin, tan, atan2, pi as M_PI

# pylint: disable=W0614,W0401
#px/from gl cimport *
from OpenGL.GL import *
#px-
from OpenGL.GL.ARB.vertex_shader import glGetActiveAttribARB as glGetActiveAttrib
#px/from glm cimport *
from .glm import *

#px/from _gldraw cimport *
from .gldraw import *
# pylint: enable=W0614,W0401

from .debug import debug, DEBUG_DRAW, DEBUG_PICK, DEBUG_MSG, DEBUG_NOVSHADER, DEBUG_NOFSHADER, DEBUG_NOSHADER

#px/DEF SOURCEGLVERSION = 'GL'
SOURCEGLVERSION = 'GL'

debug('Importing module:', __name__)
debug('  from package:', __package__)
debug('  compiled:', __compiled)
debug('  GL/GLES:', SOURCEGLVERSION)


#px/cdef struct Terrain:
class terrain: pass     # pylint: disable=W0232, C0321, R0903
    #px+float red
    #px+float green
    #px+float blue
    #px+int width
    #px+int height
#px+cdef Terrain terrain

#px/cdef struct Frustrum:
class frustrum: pass    # pylint: disable=W0232, R0903
    #px+float fovy_angle  # field of view angle
    #px+float fovy_radius
    #px+float fovy_radius_zoom
    #px+double bounding_sphere_radius
    #px+bint multisample
    #px+mat4 modelview_matrix
    #px+mat4 projection_matrix
    #px+mat4 picking_matrix
#px+cdef Frustrum frustrum

#px/cdef struct Program:
class program: pass
    #px+GLuint prog_render
    #px+GLuint projection_location
    #px+GLuint modelview_location
    #px+GLuint prog_hud  # only used if DEBUG_DRAW
    #px+GLuint prog_pick
    #px+GLuint picking_location
    #px+GLuint projection_pick_location
    #px+GLuint modelview_pick_location
#px+cdef Program program

#px/cdef struct SelectionDebugPoints:
class selection_debug_points: pass
    #px+float modelview1[3], modelview2[3]
    #px+int viewport1[2], viewport2[2]
#px+cdef SelectionDebugPoints selection_debug_points

### module state

def init_module():
    terrain.red = 0.0
    terrain.green = 0.0
    terrain.blue = 0.0
    terrain.width = 1
    terrain.height = 1
    
    frustrum.fovy_angle = 33.0  # field of view angle
    frustrum.fovy_radius = 1 / tan(frustrum.fovy_angle * M_PI / 360.0)
    frustrum.fovy_radius_zoom = frustrum.fovy_radius  # zoom == 1.
    frustrum.bounding_sphere_radius = 1.
    frustrum.multisample = 0
    # fill modelview_matrix
    frustrum.modelview_matrix = mat4(1.)
    _update_modelview_matrix_translation()
    # fill projection_matrix, see doc of glFrustum
    frustrum.projection_matrix = mat4(1.)
    frustrum.projection_matrix[2][3] = -1.
    frustrum.projection_matrix[3][3] = 0.
    _update_projection_matrix()
    # fill picking_matrix
    frustrum.picking_matrix = mat4(1.)
    
    program.prog_render = 0
    program.prog_hud = 0
    program.prog_pick = 0

#px/cdef void _update_modelview_matrix_translation():
def _update_modelview_matrix_translation():
    frustrum.modelview_matrix[3][2] = -frustrum.bounding_sphere_radius * (frustrum.fovy_radius_zoom + 1.)
    
#px/cdef void _set_modelview_matrix_rotation(float radiansx, float radiansy):
def _set_modelview_matrix_rotation(radiansx, radiansy):
    #px+cdef vec4 *M
    #px+cdef float sx, sy, cx, cy
    #px+cdef float m00, m11, m12, m20
    
    #px/M = &frustrum.modelview_matrix[0]
    M = frustrum.modelview_matrix
    
    sx = sin(radiansx/2)
    sy = sin(radiansy/2)
    cx = cos(radiansx/2)
    cy = cos(radiansy/2)
    
    m00 = 2*cx*cx - 1.
    m11 = 2*cy*cy - 1.
    m12 = 2*sy*cy
    m20 = 2*sx*cx
    # pylint: disable=C0321
    M[0][0] =  m00;         M[1][0] = 0.;   M[2][0] =  m20
    M[0][1] =  m12 * m20;   M[1][1] = m11;  M[2][1] = -m00 * m12
    M[0][2] = -m11 * m20;   M[1][2] = m12;  M[2][2] =  m00 * m11
    # py lint: enable=C0321
    
#px/cdef void _update_projection_matrix():
def _update_projection_matrix():
    if terrain.width < terrain.height:
        aspectx = 1.
        aspecty = terrain.height / terrain.width
    else:
        aspectx = terrain.width / terrain.height
        aspecty = 1.
    # see doc of glFrustum
    frustrum.projection_matrix[0][0] = frustrum.fovy_radius / aspectx
    frustrum.projection_matrix[1][1] = frustrum.fovy_radius / aspecty
    frustrum.projection_matrix[2][2] = -(frustrum.fovy_radius_zoom + 1.)
    frustrum.projection_matrix[3][2] = -(frustrum.fovy_radius_zoom + 2.
                    ) * frustrum.bounding_sphere_radius * frustrum.fovy_radius_zoom
        
#px/cdef void _set_picking_matrix(int x, int y):
def _set_picking_matrix(x, y):
    # Set picking matrix, restrict drawing to one pixel of the viewport
    # same as:  glLoadIdentity()
    #           gluPickMatrix(x, y, 1, 1, viewport)
    # same as:  glLoadIdentity()
    #           glTranslatef(terrain.width - 2*x, terrain.height - 2*y, 0.)
    #           glScalef(terrain.width, terrain.height, 1.0)
    frustrum.picking_matrix[3][0] = terrain.width - 2*x
    frustrum.picking_matrix[3][1] = terrain.height - 2*y
    frustrum.picking_matrix[0][0] = terrain.width
    frustrum.picking_matrix[1][1] = terrain.height
    
#px/cdef void _set_picking_matrix_identity():
def _set_picking_matrix_identity():
    frustrum.picking_matrix[3][0] = 0.
    frustrum.picking_matrix[3][1] = 0.
    frustrum.picking_matrix[0][0] = 1.
    frustrum.picking_matrix[1][1] = 1.
    
def set_frustrum(bounding_sphere_radius, zoom):
    frustrum.bounding_sphere_radius = bounding_sphere_radius
    frustrum.fovy_radius_zoom = frustrum.fovy_radius / zoom
    _update_modelview_matrix_translation()
    _update_projection_matrix()
    
def set_background_color(red, green, blue):
    terrain.red = red
    terrain.green = green
    terrain.blue = blue
    
def set_antialiasing(multisample):
    frustrum.multisample = multisample
    
def set_rotation_xy(x, y):
    x %= 360
    # pylint: disable=C0321
    if y < -90:     y = -90
    elif y > 90:    y = 90
    _set_modelview_matrix_rotation(M_PI * x / 180.0, M_PI * y / 180.0)
    return x, y
    
### GL state

#px/cdef void _gl_print_string(msg, GLenum name):
def _gl_print_string(msg, name):
    #px/print(msg, <char*>glGetString(name))
    print(msg, glGetString(name))
    
#px/cdef void _gl_print_float(msg, GLenum name):
def _gl_print_float(msg, name):
    #px+cdef GLfloat i
    #px/glGetFloatv(name, &i)
    i = glGetFloatv(name)
    print(msg, i)
    
#px/cdef void _gl_print_integer(msg, GLenum name):
def _gl_print_integer(msg, name):
    #px+cdef GLint i
    #px/glGetIntegerv(name, &i)
    i = glGetIntegerv(name)
    print(msg, i)
    
#px/cdef void _gl_print_bool(msg, GLenum name):
def _gl_print_bool(msg, name):
    #px+cdef GLboolean i
    #px/glGetBooleanv(name, &i)
    i = glGetBooleanv(name)
    print(msg, i)
    
def gl_init():
    if DEBUG_MSG:
        print('GL Strings:')
        _gl_print_string('  GL Vendor:', GL_VENDOR)
        _gl_print_string('  GL Renderer:', GL_RENDERER)
        _gl_print_string('  GL Version:', GL_VERSION)
        _gl_print_string('  GL Shading Language Version:', GL_SHADING_LANGUAGE_VERSION)
        #_gl_print_string('  GL Extensions:', GL_EXTENSIONS)
        _gl_print_integer('  GL_SAMPLE_BUFFERS:', GL_SAMPLE_BUFFERS)
        _gl_print_float('  GL_SAMPLE_COVERAGE_VALUE:', GL_SAMPLE_COVERAGE_VALUE)
        _gl_print_bool('  GL_SAMPLE_COVERAGE_INVERT:', GL_SAMPLE_COVERAGE_INVERT)
        _gl_print_integer('  GL_SAMPLES:', GL_SAMPLES)
        #px/IF SOURCEGLVERSION == 'GL':
        if True:
            print('  GL_MULTISAMPLE:', glIsEnabled(GL_MULTISAMPLE))
        print('  GL_SAMPLE_ALPHA_TO_COVERAGE:', glIsEnabled(GL_SAMPLE_ALPHA_TO_COVERAGE))
        #print('  GL_SAMPLE_ALPHA_TO_ONE:', glIsEnabled(GL_SAMPLE_ALPHA_TO_ONE))
        print('  GL_SAMPLE_COVERAGE:', glIsEnabled(GL_SAMPLE_COVERAGE))
    glClearColor(0., 0., 0., 1.)
    glEnable(GL_DEPTH_TEST)
    glEnable(GL_CULL_FACE) # Rasterise only the exterior faces,  to speed things up
    glCullFace(GL_BACK)
    glFrontFace(GL_CCW)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    gl_init_buffers()
    
def gl_exit():
    gl_delete_buffers()
    #px+cdef GLuint prog
    for prog in [program.prog_render, program.prog_hud, program.prog_pick]:
        if prog > 0:
            glDeleteProgram(prog)
    program.prog_render = 0
    program.prog_hud = 0
    program.prog_pick = 0
    
def gl_resize(width, height):
    terrain.width = width
    terrain.height = height
    glViewport(0, 0, terrain.width, terrain.height)
    _update_projection_matrix()
    
### render functions

#px/cdef void _gl_set_matrix(GLint location, mat4 &matrix):
def _gl_set_matrix(location, matrix):
    #px/glUniformMatrix4fv(location, 1, GL_FALSE, &matrix[0][0])
    glUniformMatrix4fv(location, 1, GL_FALSE, matrix)
    
def gl_render():
    if DEBUG_PICK:
        _set_picking_matrix_identity()
        _gl_render_pick()
    else:
        glUseProgram(program.prog_render)
        #px/IF SOURCEGLVERSION == 'GL':
        if True:
            if frustrum.multisample:
                glEnable(GL_MULTISAMPLE)
            else:
                glDisable(GL_MULTISAMPLE)
        glClearColor(terrain.red, terrain.green, terrain.blue, 1.)
        _gl_set_matrix(program.projection_location, frustrum.projection_matrix)
        _gl_set_matrix(program.modelview_location, frustrum.modelview_matrix)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        gl_draw_cube()
    if DEBUG_DRAW:
        gl_draw_cube_debug()
        
def gl_render_select_debug():
    #px+cdef GLfloat selectdata[12]
    selectdata[0] = selection_debug_points.modelview1[0]
    selectdata[1] = selection_debug_points.modelview1[1]
    selectdata[2] = selection_debug_points.modelview1[2]
    selectdata[3] = selection_debug_points.modelview2[0]
    selectdata[4] = selection_debug_points.modelview2[1]
    selectdata[5] = selection_debug_points.modelview2[2]
    selectdata[6] = selection_debug_points.viewport1[0] / terrain.width * 2 - 1
    selectdata[7] = selection_debug_points.viewport1[1] / terrain.height * 2 - 1
    selectdata[8] = 0
    selectdata[9] = selection_debug_points.viewport2[0] / terrain.width * 2 - 1
    selectdata[10] = selection_debug_points.viewport2[1] / terrain.height * 2 - 1
    selectdata[11] = 0
    #px/gl_draw_select_debug(&selectdata[0], sizeof(selectdata), program.prog_hud)
    gl_draw_select_debug(selectdata, sizeof(selectdata), program.prog_hud)
    

### picking functions

#px/cdef void _gl_render_pick():
def _gl_render_pick():
    glUseProgram(program.prog_pick)
    #px/IF SOURCEGLVERSION == 'GL':
    if True:
        glDisable(GL_MULTISAMPLE)
    glClearColor(0., 0., 0., 1.)
    _gl_set_matrix(program.picking_location, frustrum.picking_matrix)
    _gl_set_matrix(program.projection_pick_location, frustrum.projection_matrix)
    _gl_set_matrix(program.modelview_pick_location, frustrum.modelview_matrix)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    gl_pick_cube()
    
#px/cpdef gl_pick_polygons(int x, int y):
def gl_pick_polygons(x, y):
    #px+cdef unsigned char pixel[3]
    
    if not (0 <= x < terrain.width and 0 <= y < terrain.height):
        return 0
    _set_picking_matrix(x, y)
    _gl_render_pick()
    #px/glReadPixels(x, y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, pixel)
    pixel = glReadPixels(x, y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, [[[0, 0, 0]]])[0][0]
    index = pixel[0]<<4 | pixel[1] | pixel[2]>>4
    return index
    
#px/cdef void _modelview_to_viewport(vvect, int *mvect):
def _modelview_to_viewport(vvect, mvect):
    #px+cdef vec4 *M
    #px+cdef vec4 *P
    #px+cdef float u0, u1, u2, v0, v1, v3
    
    #px/M = &frustrum.modelview_matrix[0]
    M = frustrum.modelview_matrix
    #px/P = &frustrum.projection_matrix[0]
    P = frustrum.projection_matrix
    
    # u = M^T * vvect
    u0 = M[0][0]*vvect[0] + M[1][0]*vvect[1] + M[2][0]*vvect[2] + M[3][0]
    u1 = M[0][1]*vvect[0] + M[1][1]*vvect[1] + M[2][1]*vvect[2] + M[3][1]
    u2 = M[0][2]*vvect[0] + M[1][2]*vvect[1] + M[2][2]*vvect[2] + M[3][2]
    #u3 = 1.
    
    # v = P * u
    v0 = P[0][0] * u0 + P[1][0] * u1 + P[2][0] * u2 + P[3][0] #* u3
    v1 = P[0][1] * u0 + P[1][1] * u1 + P[2][1] * u2 + P[3][1] #* u3
    #v2 = P[0][2] * u0 + P[1][2] * u1 + P[2][2] * u2 + P[3][2] * u3
    v3 = P[0][3] * u0 + P[1][3] * u1 + P[2][3] * u2 + P[3][3] #* u3
    
    mvect[0] = int((v0 / v3 + 1) / 2 * terrain.width)
    mvect[1] = int((v1 / v3 + 1) / 2 * terrain.height)
    
def get_cursor_angle(point1, point2):
    '''The result is the angle (on the screen)
    at which the mouse cursor needs to be drawn.'''
    #px+cdef float angle
    #px+cdef int i, x, y
    for i in range(3):
        selection_debug_points.modelview1[i] = point1[i]
        selection_debug_points.modelview2[i] = point2[i]
    _modelview_to_viewport(point1, selection_debug_points.viewport1)
    _modelview_to_viewport(point2, selection_debug_points.viewport2)
    x = selection_debug_points.viewport1[0] - selection_debug_points.viewport2[0]
    y = selection_debug_points.viewport1[1] - selection_debug_points.viewport2[1]
    angle = atan2(x, y) * 180.0 / M_PI
    return angle
    
    
### shader functions

#px/cdef void _gl_print_shader_log(GLuint shader):
def _gl_print_shader_log(shader):
    #px+cdef GLint log_len
    #px+cdef GLsizei length
    #px+cdef char log[1024]
    #px/glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &log_len)
    log_len = glGetShaderiv(shader, GL_INFO_LOG_LENGTH)
    if log_len > 0:
        print('==== Error compiling shader:')
        #px/glGetShaderInfoLog(shader, 1023, &length, log)
        log = glGetShaderInfoLog(shader, 1023)
        print(log.decode('utf-8').rstrip())
        print('====')
    else:
        print('==== Error compiling shader (no log)')
        
#px/cdef void _gl_print_program_log(GLuint program):
def _gl_print_program_log(program):
    #px+cdef GLint log_len
    #px+cdef GLsizei length
    #px+cdef char log[1024]
    #px/glGetProgramiv(program, GL_INFO_LOG_LENGTH, &log_len)
    log_len = glGetProgramiv(program, GL_INFO_LOG_LENGTH)
    if log_len > 0:
        print('==== Error linking shader program:')
        #px/glGetProgramInfoLog(program, 1023, &length, log)
        log = glGetProgramInfoLog(program, 1023)
        print(log.decode('utf-8').rstrip())
        print('====')
    else:
        print('==== Error linking shader program (no log)')
        
#px/cdef GLuint _gl_create_compiled_shader(GLenum shadertype, bytes source):
def _gl_create_compiled_shader(shadertype, source):
    #px+cdef GLuint shader
    #px+cdef const_GLchar_ptr pchar
    #px+cdef GLint compile_status
    
    shader = glCreateShader(shadertype)
    if shader == 0:
        print('Failed to create shader')
        return 0
    #px+pchar = <const_GLchar_ptr><char*>source
    #px/glShaderSource(shader, 1, &pchar, NULL)
    glShaderSource(shader, source)
    glCompileShader(shader)
    #px/glGetShaderiv(shader, GL_COMPILE_STATUS, &compile_status)
    compile_status = glGetShaderiv(shader, GL_COMPILE_STATUS)
    if not compile_status:
        _gl_print_shader_log(shader)
        return 0
    return shader
    
#px/cdef void _gl_print_program_info(GLuint program):
def _gl_print_program_info(program):
    #px+cdef GLint param
    #px+cdef int i
    #px+cdef GLsizei alength
    #px+cdef GLint asize, location
    #px+cdef GLenum atype
    #px+cdef char aname[1024]
    
    print('shader program info', program)
    glValidateProgram(program)
    program_info = {}
    for msg, pname in [('delete status', GL_DELETE_STATUS),
                       ('link status', GL_LINK_STATUS),
                       ('validate status', GL_VALIDATE_STATUS),
                       ('info log length', GL_INFO_LOG_LENGTH),
                       ('attached shaders', GL_ATTACHED_SHADERS),
                       ('active attributes', GL_ACTIVE_ATTRIBUTES),
                       ('active attribute max length', GL_ACTIVE_ATTRIBUTE_MAX_LENGTH),
                       ('active uniforms', GL_ACTIVE_UNIFORMS),
                       ('active uniform max length', GL_ACTIVE_UNIFORM_MAX_LENGTH)]:
        #px/glGetProgramiv(program, pname, &param)
        param = glGetProgramiv(program, pname)
        program_info[pname] = param
        print(' ', msg, param)
    print('active attributes:')
    for i in range(program_info[GL_ACTIVE_ATTRIBUTES]):
        #px/glGetActiveAttrib(program, i, 1023, &alength, &asize, &atype, aname)
        aname, asize, atype = glGetActiveAttrib(program, i); alength = '?'
        location = glGetAttribLocation(program, aname)
        print('  {}, {}: length={} size={} type={} location={}'.format(i, aname, alength, asize, atype, location))
    print('active uniforms:')
    for i in range(program_info[GL_ACTIVE_UNIFORMS]):
        #px/glGetActiveUniform(program, i, 1023, &alength, &asize, &atype, aname)
        aname, asize, atype = glGetActiveUniform(program, i); alength = '?'
        location = glGetUniformLocation(program, aname)
        print('  {}, {}: length={} size={} type={} location={}'.format(i, aname, alength, asize, atype, location))
        
#px/cdef GLuint _gl_create_program(bytes vertex_source, bytes fragment_source, list attributes):
def _gl_create_program(vertex_source, fragment_source, attributes):
    #px+cdef GLuint vertex_shader = 0, fragment_shader = 0
    #px+cdef GLuint program
    #px+cdef GLint link_status
    
    if DEBUG_NOSHADER:
        return 0
        
    program = glCreateProgram()
    if not DEBUG_NOVSHADER:
        debug('  creating vertex shader')
        vertex_shader = _gl_create_compiled_shader(GL_VERTEX_SHADER, vertex_source)
        glAttachShader(program, vertex_shader)
    if not DEBUG_NOFSHADER:
        debug('  creating fragment shader')
        fragment_shader = _gl_create_compiled_shader(GL_FRAGMENT_SHADER, fragment_source)
        glAttachShader(program, fragment_shader)
    for index, name in attributes:
        glBindAttribLocation(program, index, name)
    glLinkProgram(program)
    #px/glGetProgramiv(program, GL_LINK_STATUS, &link_status)
    link_status = glGetProgramiv(program, GL_LINK_STATUS)
    if not link_status:
        _gl_print_program_log(program)
        return 0
    if DEBUG_MSG:
        _gl_print_program_info(program)
    glDetachShader(program, vertex_shader)
    glDetachShader(program, fragment_shader)
    glDeleteShader(vertex_shader)
    glDeleteShader(fragment_shader)
    return program
    
#px/def gl_create_render_program(bytes vertex_source, bytes fragment_source):
def gl_create_render_program(vertex_source, fragment_source):
    #px+cdef GLint location
    if program.prog_render > 0:
        glDeleteProgram(program.prog_render)
    attributes = [
                  (VERTEX_ATTRIB_LOCATION, b'vertex_attr'),
                  (NORMAL_ATTRIB_LOCATION, b'normal_attr'),
                  (COLOR_ATTRIB_LOCATION, b'color_attr'),
                  (TEXCOORD_ATTRIB_LOCATION, b'texcoord_attr'),
                 ]
    program.prog_render = _gl_create_program(vertex_source, fragment_source, attributes)
    if program.prog_render > 0:
        glUseProgram(program.prog_render)
        location = glGetUniformLocation(program.prog_render, b'tex')
        glUniform1i(location, 0) # 0 is the texture unit (-> glActiveTexture)
        program.modelview_location = glGetUniformLocation(program.prog_render, b'modelview')
        program.projection_location = glGetUniformLocation(program.prog_render, b'projection')
        location = glGetUniformLocation(program.prog_render, b'object')
        gl_init_object_location(location)
        
#px/def gl_create_hud_program(bytes vertex_source, bytes fragment_source):
def gl_create_hud_program(vertex_source, fragment_source):
    #px+cdef GLint location
    if program.prog_hud > 0:
        glDeleteProgram(program.prog_hud)
    attributes = [
                  (VERTEX_ATTRIB_LOCATION, b'vertex_attr'),
                  (COLOR_ATTRIB_LOCATION, b'color_attr'),
                 ]
    program.prog_hud = _gl_create_program(vertex_source, fragment_source, attributes)
    if program.prog_hud > 0:
        glUseProgram(program.prog_hud)
    
#px/def gl_create_pick_program(bytes vertex_source, bytes fragment_source):
def gl_create_pick_program(vertex_source, fragment_source):
    if program.prog_pick > 0:
        glDeleteProgram(program.prog_pick)
    attributes = [(VERTEX_ATTRIB_LOCATION, b'vertex_attr'),
                  (COLOR_ATTRIB_LOCATION, b'color_attr')]
    program.prog_pick = _gl_create_program(vertex_source, fragment_source, attributes)
    if program.prog_pick > 0:
        glUseProgram(program.prog_pick)
        program.picking_location = glGetUniformLocation(program.prog_pick, b'picking')
        program.projection_pick_location = glGetUniformLocation(program.prog_pick, b'projection')
        program.modelview_pick_location = glGetUniformLocation(program.prog_pick, b'modelview')
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.



import sys, os

from PyQt4.QtCore import Qt, QPoint
from PyQt4.QtGui import QDialogButtonBox
from PyQt4.QtTest import QTest


field_functions = {
        'view_editbar':     ('action_editbar', False, lambda widget: widget.isChecked()),
        'view_toolbar':     ('action_toolbar', False, lambda widget: widget.isChecked()),
        'view_statusbar':   ('action_statusbar', False, lambda widget: widget.isChecked()),
        'edit_text':        ('edit_moves', None, lambda widget: str(widget.text())),
        'edit_pos':         ('edit_moves', None, lambda widget: widget.cursorPosition()),
        'game_pos':         ('MainWindow', None, lambda widget: widget.gamestate.all_moves.current_place),
        'game_len':         ('MainWindow', None, lambda widget: len(widget.gamestate.all_moves.moves)),
        'solved':           ('MainWindow', None, lambda widget: widget.gamestate.current_cube_state.is_solved()),
        'sizes':            ('MainWindow', None, lambda widget: widget.gamestate.current_cube_state.model.sizes),
        'rotationx':        ('MainWindow', None, lambda widget: widget.cube_area.rotation_x),
        'rotationy':        ('MainWindow', None, lambda widget: widget.cube_area.rotation_y),
        'selectdlg_exists': ('DialogSelectModel', False, lambda widget: True),
        'selectdlg_visible':('DialogSelectModel', False, lambda widget: widget.isVisible()),
        'dlg_model':        ('combobox_model', None, lambda widget: widget.currentIndex()),
        'lblsize1':         ('label_width', None, lambda widget: widget.isEnabled()),
        'lblsize2':         ('label_heigth', None, lambda widget: widget.isEnabled()),
        'lblsize3':         ('label_depth', None, lambda widget: widget.isEnabled()),
        'size1':            ('spin_size1', None, lambda widget: (widget.value(), widget.isEnabled())),
        'size2':            ('spin_size2', None, lambda widget: (widget.value(), widget.isEnabled())),
        'size3':            ('spin_size3', None, lambda widget: (widget.value(), widget.isEnabled())),
        'dlg_solved':       ('checkbox_solved', None, lambda widget: widget.isChecked()),
        'preferencesdlg_exists': ('DialogPreferences', False, lambda widget: True),
        'preferencesdlg_visible': ('DialogPreferences', False, lambda widget: widget.isVisible()),
        'aboutdlg_exists': ('AboutDialog', False, lambda widget: True),
        'aboutdlg_visible': ('AboutDialog', False, lambda widget: widget.isVisible()),
    }

widget_functions = []

def widget(*wnames, **kwargs):
    fname = kwargs.setdefault('fname', None)
    assert list(kwargs.keys()) == ['fname'], kwargs
    def decorator(func):
        widget_functions.append((wnames, fname or func.__name__, func))
        return func
    return decorator
    
    
@widget('edit_moves')
def edit_moves_text(widget, text, enter=True):
    widget.setText(text)
    if enter:
        widget.returnPressed.emit()
        
edit_moves_key = widget('edit_moves', fname='edit_moves_key')(QTest.keyClick)

@widget('edit_moves')
def edit_moves_key_enter(widget, key, modifiers=Qt.NoModifier):
    QTest.keyClick(widget, key, modifiers)
    QTest.keyClick(widget, Qt.Key_Enter)
    
button_edit_exec_click = widget('button_edit_exec', fname='button_edit_exec_click')(QTest.mouseClick)
button_edit_clear_click = widget('button_edit_clear', fname='button_edit_clear_click')(QTest.mouseClick)

drawingarea_key = widget('drawingarea', fname='drawingarea_key')(QTest.keyClick)

@widget('drawingarea')
def drawingarea_mouse_click(widget, p, button, modifiers=Qt.NoModifier):
    QTest.mouseMove(widget, QPoint(*p))
    QTest.mouseClick(widget, button, modifiers, QPoint(*p), delay=50)
    
@widget('edit_moves', 'drawingarea')
def empty_and_drawingarea_mouse_click(edit_moves, drawingarea, p, button, modifiers=Qt.NoModifier):
    edit_moves_text(edit_moves, '')
    QTest.qWait(50)
    drawingarea_mouse_click(drawingarea, p, button, modifiers)
    
@widget('drawingarea')
def drawingarea_mouse_move(widget, p1, p2, button, modifiers=Qt.NoModifier):
    QTest.mouseMove(widget, QPoint(*p1))
    QTest.mousePress(widget, button, modifiers, QPoint(*p1), delay=50)
    QTest.mouseMove(widget, QPoint(*p2))
    QTest.mouseRelease(widget, button, modifiers, QPoint(*p2))
    
@widget('action_selectmodel')
def dialog_selectmodel(action):
    action.trigger()
    QTest.qWait(200)
    return 'find_qobjects'
    
@widget('combobox_model')
def dialog_selectmodel_changemodel(widget, key):
    if widget.isVisible():
        QTest.keyClick(widget, key, delay=100)
        
@widget('checkbox_solved')
def dialog_selectmodel_checksolved(widget, check):
    if widget.isVisible():
        QTest.qWait(100)
        widget.setChecked(check)
    
@widget('buttonBox')
def dialog_selectmodel_ok(widget):
    if widget.isVisible():
        QTest.qWait(100)
        QTest.mouseClick(widget.button(QDialogButtonBox.Ok), Qt.LeftButton)
        QTest.qWait(100)
    
@widget('buttonBox')
def dialog_selectmodel_cancel(widget):
    if widget.isVisible():
        QTest.qWait(100)
        QTest.mouseClick(widget.button(QDialogButtonBox.Cancel), Qt.LeftButton)
        QTest.qWait(100)
    
@widget('action_preferences')
def dialog_preferences(action):
    action.trigger()
    QTest.qWait(200)
    return 'find_qobjects'
    
@widget('buttonBox')
def dialog_preferences_close(widget):
    if widget.isVisible():
        QTest.qWait(100)
        QTest.mouseClick(widget.button(QDialogButtonBox.Close), Qt.LeftButton)
        QTest.qWait(100)
        
#TODO: The testrunner can't currently test modal dialogs, this function will block
@widget('action_info')
def dialog_about(action):
    action.trigger()
    QTest.qWait(200)
    return 'find_qobjects'
    
@widget('buttonBox')
def dialog_about_close(widget):
    if widget.isVisible():
        QTest.qWait(100)
        QTest.mouseClick(widget.button(QDialogButtonBox.Close), Qt.LeftButton)
        QTest.qWait(100)
        





#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import random
from copy import deepcopy

from .debug import *    # pylint: disable=W0614,W0401


# This class is currently used to store the initial state of the cube
# and to provide the cube to plugins.
class CubeState (object):
    rand = random.Random()
    if DEBUG_RAND:
        print('rand.seed(123)')
        rand.seed(123)
    
    def __init__(self, model):
        self.model = model
        self._blocksn = None
        
    def copy(self):
        other = CubeState(self.model)
        other._blocksn = deepcopy(self._blocksn)
        return other
        
    @property
    def type(self):
        return self.model.type
    @property
    def size(self):
        return self.model.size
    @property
    def blocks(self):
        assert self._blocksn is not None
        blocks = [self.model.rotation_symbolic_to_matrix(*blockn)
                        for blockn in self._blocksn]
        return blocks
    
    def set_model(self, model):
        assert self.model.type == model.type
        assert self.model.sizes == model.sizes
        self.model = model
        
    def set_solved(self):
        # '' symbolic rotation for identity
        self._blocksn = [(i, '') for i, unused_block in enumerate(self.model.blocks)]
        
    def is_solved(self):
        '''Check whether the cube is solved.
        
        Only test the colors of the visible faces.
        Cubes with rotated center faces are treated as solved.
        '''
        face_colors = {}
        for index, (unused_pos, rot) in enumerate(self._blocksn):
            for face in self.model.blocks[index].visible_faces:
                color = self.get_colorsymbol(index, face.lower())
                try:
                    color_ref = face_colors[face]
                except KeyError:
                    face_colors[face] = color
                else:
                    if color != color_ref:
                        return False
        return True
        
    def is_solved_strict(self):
        '''Check whether the cube is solved.
        
        Test the rotation and position of the blocks.
        Cubes with rotated center faces are treated as not solved.
        '''
        #FIXME: currently unused function, should be used for faces with images
        allrot = self._blocksn[0][1]
        for index, (pos, rot) in enumerate(self._blocksn):
            if rot != allrot:
                return False  # Cubie rotated
            block = self.model.rotated_position[index, rot]
            if pos != block:
                return False  # Cubie at wrong place
        return True
        
    def identify_rotation_blocks(self, maxis, mslice):
        assert self._blocksn
        if mslice == -1:
            for i in range(len(self._blocksn)):
                yield i, True
        else:
            for i, blockn in enumerate(self._blocksn):
                bslice = self.model.blocks[blockn[0]].axis_to_slice(maxis)
                yield i, bslice == mslice
                
    def _rotate_slice(self, axis, slice_, dir_):
        assert self._blocksn
        
        if DEBUG_ROTATE:
            print('rotate axis={} slice={} dir={!s:5}\n  blocks:'.format(axis, slice_, dir_), end=' ')
        for block_id, selected in self.identify_rotation_blocks(axis, slice_):
            if  selected:
                if DEBUG_ROTATE:
                    print('{}:{}{}'.format(block_id, *self._blocksn[block_id]), end='')
                block = self.model.rotate_symbolic(axis, dir_, *self._blocksn[block_id])
                if DEBUG_ROTATE:
                    print('-{}{}'.format(*block), end=' ')
                self._blocksn[block_id] = block
        if DEBUG_ROTATE:
            print()
        
    def get_colorsymbol(self, blocknum, facesym):
        for pos, rot in self._blocksn:
            if pos == blocknum:
                return self.model.face_symbolic_to_face_color(facesym, rot)
        assert False, 'Should not be reached'
        
    def get_colornum(self, blocknum, facesym):
        #FIXME: face and color symbols should be upper
        colorsym = self.get_colorsymbol(blocknum, facesym.lower())
        return self.model.faces.index(colorsym.upper())
        
    def format_block(self):
        assert self._blocksn is not None
        # every block is stored as pos-sym, where sym is a symbolic rotation
        blocks = ['{}-{}'.format(pos, sym) for pos, sym in self._blocksn]
        return 'idx-rot: ' + ' '.join(blocks)
        
    def parse_block(self, blocks):
        if blocks == 'solved':
            return self.set_solved()
        bformat, blocks = blocks.split(':', 1)
        if bformat != 'idx-rot':
            raise ValueError('unknown block format:', bformat)
        blocks = blocks.strip().split(' ')
        if len(blocks) != len(self.model.blocks):
            raise ValueError('wrong block count: %s, expected: %s' % (len(blocks), len(self.model.blocks)))
        blocksn = []
        for block in blocks:
            # every block is stored as idx-rot, where idx: index to blocklist, rot: symbolic rotation
            block = block.strip().split('-', 1)
            index, rot = block
            index = int(index)
            rot = self.model.norm_symbol(rot)
            blocksn.append((index, rot))
        # test whether block indices is a permutation,
        # in fact thats not enough, e.g. swap a corner cubie with an edge,
        # also cubie rotation may be invalid, it can be possible that a
        # label is rotated inside the cube.
        for i1, i2 in enumerate(sorted(i for i, r in blocksn)):
            if i1 != i2:
                raise ValueError('block list is not a permutation')
        self._blocksn = blocksn
        
    def randomize(self, count=-1):
        if not sum(self.model.sizes):
            return
        if count >= 0:
            midslice = -1
            count_ = count
        else:
            midslice = [s // 2 for s in self.model.sizes]
            count_ = 20 * sum(self.model.sizes)
        cubemoves = []
        debug('random moves:', count_)
        lastaxis = None
        lastslices = []
        for unused_i in range(count_):
            while True:
                maxis = self.rand.randrange(3)
                mdir = bool(self.rand.randrange(2))
                msign = -1 if mdir else 1
                mslice = self.rand.randrange(self.model.sizes[maxis])
                if maxis != lastaxis:
                    lastaxis = maxis
                    lastslices = [0] * self.model.sizes[maxis]
                    lastslices[mslice] = msign
                    break # subsequent different axes are accepted
                nslices = lastslices[mslice] + msign
                if abs(nslices) < abs(lastslices[mslice]):
                    continue # moves that revert a previous one are rejected
                if abs(nslices * self.model.symmetry[maxis]) > 180.:
                    continue # moves that rotate a slice more than 180° are rejected
                lastslices[mslice] += msign
                if (sum(1 for ls in lastslices if ls > 0 or ls * self.model.symmetry[maxis] == -180.)
                        > self.model.sizes[maxis] // 2):
                    lastslices[mslice] -= msign
                    continue
                if (sum(1 for ls in lastslices if ls < 0 or ls * self.model.symmetry[maxis] == 180.)
                        > self.model.sizes[maxis] // 2):
                    lastslices[mslice] -= msign
                    continue
                break
            if mslice == midslice:
                cubemoves.append((maxis, mdir))
            self._rotate_slice(maxis, mslice, mdir)
        for maxis, mdir in reversed(cubemoves):
            self._rotate_slice(maxis, -1, not mdir)
            
    def rotate_slice(self, move_data):
        self._rotate_slice(*move_data)
        
    def swap_block(self, blockpos1, maxis, mslice, mdir):
        if DEBUG_ROTATE:
            print('rotate axis={} slice={} dir={!s:5}\n  from:'.format(maxis, mslice, mdir), end=' ')
            
        for blockidx1, (pos, blockrot1) in enumerate(self._blocksn):
            if pos == blockpos1:
                break
        else:
            blockidx1 = blockrot1 = None
            assert False
        blockpos1r, blockrot1r = self.model.rotate_symbolic(maxis, mdir, blockpos1, blockrot1)
        if DEBUG_ROTATE:
            print('{}:{}{}->{}{}\n    to:'.format(blockidx1, blockpos1, blockrot1,
                                                 blockpos1r, blockrot1r), end=' ')
        for blockidx2, unused_selected in self.identify_rotation_blocks(maxis, mslice):
            blockpos2, blockrot2 = self._blocksn[blockidx2]
            if blockpos1r == blockpos2:
                self._blocksn[blockidx1] = blockpos1r, blockrot1r
                blockpos2r, blockrot2r = self.model.rotate_symbolic(maxis, not mdir, blockpos2, blockrot2)
                if DEBUG_ROTATE:
                    print('{}:{}{}->{}{}'.format(blockidx2, blockpos2, blockrot2,
                                                blockpos2r, blockrot2r), end=' ')
                self._blocksn[blockidx2] = blockpos2r, blockrot2r
                break
        if DEBUG_ROTATE:
            print()
        
    def rotate_block(self, blockpos, rdir):
        for blockidx, (pos, blockrot) in enumerate(self._blocksn):
            if pos == blockpos:
                break
        else:
            blockidx = blockrot = None
            assert False
        block = self.model.blocks[blockpos]
        try:
            rot = block.inplace_rotations[-1 if rdir else 0]
        except IndexError:
            # not every block can be rotated inline: e.g. edges and center faces on the 4×4×4-Cube,
            # edges and corners on towers and bricks
            #TODO: swap edge at pos n with the one at pos (size-1 - n),
            #      rotate all center faces on the same ring
            return
        blockrot2 = self.model.norm_symbol(blockrot + rot)
        self._blocksn[blockidx] = blockpos, blockrot2
        if DEBUG_ROTATE:
            sym1, colorsym1 = self.model.block_index_to_block_symbolic(blockpos, blockrot)
            sym2, colorsym2 = self.model.block_index_to_block_symbolic(blockpos, blockrot2)
            print('{}:{}{}->{}{} ({}:{}->{}:{})'.format(blockidx, blockpos, blockrot, blockpos, blockrot2,
                                                      sym1, colorsym1, sym2, colorsym2))
            self.debug_blocksymbols(allsyms=False)
            
    def debug_blocksymbols(self, allsyms):
        for blockpos, blockrot in self._blocksn:
            blocksym, colorsym = self.model.block_index_to_block_symbolic(blockpos, blockrot)
            if allsyms or blocksym != colorsym:
                print(' {}:{}'.format(blocksym, colorsym), end='')
        print('')
        
        



# -*- coding: utf-8 -*-

#  Copyright © 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


def po_isempty(pofilename, verbose=True):
    try:
        import polib
    except ImportError:
        if verbose:
            print('polib not found, assuming the file %s contains translations' % pofilename)
        return False
    for entry in polib.pofile(pofilename):
        if entry.translated() and entry.msgid != entry.msgstr:
            return False
    return True
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


def convert(salg):
    moves = {
        "": "",
        "F":   "f",    "B":   "b",    "R":   "r",    "L":  "l",    "U":   "u",    "D":   "d",
        "F'":  "f-",   "B'":  "b-",   "R'":  "r-",   "L'": "l-",   "U'":  "u-",   "D'":  "d-",
        "F2":  "ff",   "B2":  "bb",   "R2":  "rr",   "L2": "ll",   "U2":  "uu",   "D2":  "dd",
        "F'2": "ff",   "B'2": "bb",   "R'2": "rr",   "L'2": "ll",  "U'2": "uu",   "D'2": "dd",
        "M":   "r2-",  "E":   "u2-",  "S":   "f2",
        "M'":  "r2",   "E'":  "u2",   "S'":  "f2-",
        "M2":  "r2r2", "E2":  "u2u2", "S2":  "f2f2",
        "M2'": "r2r2", "E2'": "u2u2", "S2'": "f2f2",
        "x":   "R",   "y":   "U",   "z":   "F",
        "x'":  "R-",  "y'":  "U-",  "z'":  "F-",
        "x2":  "RR",  "y2":  "UU",  "z2":  "FF",
        "x2'": "RR",  "y2'": "UU",  "z2'": "FF",
    }
    palg = ''
    for m in salg.split(' '):
        palg += moves[m.strip('()')]
    return palg
    
def main():
    while True:
        try:
            line = input('> ')
        except EOFError:
            return
        print(':', convert(line))
        
        
if __name__ == '__main__':
    main()
    




#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This line is not really needed, but leave it until glarea and gldraw are compiled with cython3
from __future__ import print_function, division

import sys


DEBUG = False
DEBUG_FUNC = DEBUG_MSG = DEBUG_RAND = DEBUG_ROTATE = False
DEBUG_DRAW = DEBUG_KEYS = DEBUG_PICK = False
DEBUG_FPS = DEBUG_VFPS = DEBUG_NOLABEL = DEBUG_NOBEVEL = DEBUG_INNER = False
DEBUG_ALG = DEBUG_NOVSHADER = DEBUG_NOFSHADER = DEBUG_NOSHADER = False
DEBUG_THREAD = DEBUG_PUREPYTHON = False

__all__ = ['error', 'debug', 'debug_func', 'DEBUG',
          ] + [__n for __n in dir(sys.modules[__name__]) if __n.startswith('DEBUG_')]

def error(*args, **kwargs):
    print('ERROR:', *args, **kwargs)
    
debug = lambda *args, **kwargs: None
debug_func = lambda x: x

def set_flags(debug_flags):
    module = sys.modules[__name__]
    module.DEBUG = True
    
    for flag in debug_flags:
        setattr(module, 'DEBUG_' + flag.upper(), True)
        
    if module.DEBUG_VFPS:
        module.DEBUG_FPS = True
    
    if module.DEBUG_FUNC:
        def _debug_pre_func(func, *args, **kwargs):
            def short_arg(arg):
                arg = str(arg)
                maxlen = 80 - debug_func.indent * 3 - 2
                if len(arg) > maxlen:
                    return arg[:maxlen]
                return arg
            try:
                func_name = func.__name__
            except AttributeError:
                func_name = func.__name__
            print('%s--%s' % ('  |'*debug_func.indent, func_name))
            debug_func.indent += 1
            for arg in args:
                try:
                    print('%s: %s' % ('  |'*debug_func.indent, short_arg(arg)))
                except Exception:
                    pass
            for kw, arg in kwargs:
                try:
                    print('%s: %s=%s' % ('  |'*debug_func.indent, kw, short_arg(arg)))
                except Exception:
                    pass
        def debug_func(func):
            def ret_func(*args, **kwargs):
                _debug_pre_func(func, *args, **kwargs)
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if debug_func.last_exc != e:
                        print('  X'*debug_func.indent)
                        debug_func.last_exc = e
                    raise
                finally:
                    debug_func.indent -= 1
                    print('  |'*debug_func.indent + "--'")
            try:
                ret_func.__dict__ = func.__dict__
            except AttributeError:
                pass
            ret_func.__doc__ = func.__doc__
            ret_func.__module__ = func.__module__
            ret_func.__name__ = func.__name__
            try:
                ret_func.__defaults__ = func.__defaults__
            except AttributeError:
                pass
            return ret_func
        debug_func.indent = 0
        debug_func.last_exc = None
        module.debug_func = debug_func
        
    if module.DEBUG_MSG:
        module.debug = print
        



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
import itertools
from collections import OrderedDict
from xml.etree import ElementTree


def get_string(prop):
    string = prop.find('string')
    if string is None or string.attrib.get('notr') == 'true':
        return None
    return string.text
    
def parse_file(filename):
    tree = ElementTree.parse(filename)
    root = tree.getroot()
    window = root.find('widget')
    widgets = OrderedDict()
    child_items = []
    
    for prop in window.findall('property'):
        text = get_string(prop)
        if text is not None:
            child_items.append(('window', prop.attrib['name'], text))
    for tag in window:
        for widget in itertools.chain(tag.iter('widget'), tag.iter('action')):
            if widget.tag == 'widget':
                widget_class = widget.attrib['class']
            else:
                widget_class = 'QAction'
            widget_name = widget.attrib['name']
            for prop in itertools.chain(widget.findall('property'), widget.findall('attribute')):
                itemtype = prop.tag
                prop_name = prop.attrib['name']
                text = get_string(prop)
                if text is None:
                    continue
                if itemtype == 'property':
                    child_items.append((itemtype, widget_name, prop_name, text))
                if itemtype == 'attribute':
                    parent = window.find(".//*[@name='%s']/.." % widget_name)
                    if parent and parent.attrib.get('class') == 'QTabWidget':
                        parent_name = parent.attrib['name']
                        widgets.setdefault(parent_name, 'QTabWidget')
                        child_items.append(('QTabWidget', parent_name, widget_name, text))
                    else:
                        child_items.append((itemtype, widget_class, widget_name, prop_name, text))
                widgets.setdefault(widget_name, widget_class)
    return window.attrib['class'], widgets, child_items
    
def write_file(window_class, widgets, child_items, file):
    print(
'''# -*- coding: utf-8 -*-
# generated with: {}

from PyQt4.QtGui import {}

def retranslate(window):'''.format(' '.join(sys.argv), ', '.join(sorted(set(widgets.values())))), file=file)
    
    for widget_name, widget_class in widgets.items():
        print('    {1} = window.findChild({0}, {1!r})'.format(widget_class, widget_name), file=file)
    print(file=file)
    
    for itemtype, *itemargs in child_items:
        if itemtype == 'window':
            prop_name, text = itemargs
            prop_name = ''.join(('set', prop_name[0].upper(), prop_name[1:]))
            print('    window.{}(_({!r}))'.format(prop_name, text), file=file)
        elif itemtype == 'property':
            print('    {0}.setProperty({1!r}, _({2!r}))'.format(*itemargs), file=file)
        elif itemtype == 'QTabWidget':
            print('    {0}.setTabText({0}.indexOf({1}), _({2!r}))'.format(*itemargs), file=file)
        else:
            print('    #unknown:', itemtype, *itemargs)
        
def usage(exitcode):
    print('usage:')
    print(' ', os.path.basename(__file__), '-h|--help')
    print(' ', os.path.basename(__file__), 'uifilename')
    print(' ', os.path.basename(__file__), 'uifilename', 'pyfilename')
    sys.exit(exitcode)
    
def main():
    args = sys.argv[1:]
    if not args:
        usage(1)
    if args[0] in ['-h', '--help']:
        usage(0)
    windowinfo = parse_file(args[0])
    args = args[1:]
    if args:
        with open(args[0], 'wt', encoding='utf-8') as file:
            write_file(*windowinfo, file=file)
    else:
        write_file(*windowinfo, file=sys.stdout)
        
        
if __name__ == '__main__':
    main()
    




#-*- coding:utf-8 -*-

#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=C0321


def play_stripes_Brick_nnn_2(game):
    size = game.current_state.model.sizes
    for unused_j in 0, 1:
        for i in range(size[1]-1, size[1]//2-1, -1):   game.add_moves([(1, i, 0)])
        for i in range(1, size[0], 2):                 game.add_moves([(0, i, 0)])
        for i in range(1, size[2], 2):                 game.add_moves([(2, i, 0)])
        
def play_stripes_Brick_nn1_4(game):
    size = game.current_state.model.sizes
    for i in range(1, size[0], 2):                 game.add_moves([(0, i, 0)])
    for i in range(size[1]-1, size[1]//2, -1):     game.add_moves([(1, i, 0)])
    for i in range(1, size[0] // 2, 2):            game.add_moves([(0, i, 0)])
    game.add_moves([(1, size[1]//2, 0)])
    for i in range(size[0]//2+1, size[0], 2):      game.add_moves([(0, i, 0)])
    for i in range(size[1]-1, size[1]//2-1, -1):   game.add_moves([(1, i, 0)])
    
def play_stripes_Brick_1nn_4(game):
    size = game.current_state.model.sizes
    for i in range(1, size[2], 2):                 game.add_moves([(2, i, 0)])
    for i in range(size[1]-1, size[1]//2, -1):     game.add_moves([(1, i, 0)])
    for i in range(1, size[2] // 2, 2):            game.add_moves([(2, i, 0)])
    game.add_moves([(1, size[1]//2, 0)])
    for i in range(size[2]//2+1, size[2], 2):      game.add_moves([(2, i, 0)])
    for i in range(size[1]-1, size[1]//2-1, -1):   game.add_moves([(1, i, 0)])
    
def play_stripes_Cube_nnn_2(game):
    size = game.current_state.model.sizes
    for i in range(0, size[0], 2): game.add_moves([(0, i, 0)] * 2)
    for i in range(0, size[2], 2): game.add_moves([(2, i, 0)] * 2)
    for i in range(0, size[0], 2): game.add_moves([(0, i, 0)] * 2)
    
def play_stripes_TowerBrick_nnn_2(game):
    size = game.current_state.model.sizes
    for i in range(0, size[0], 2): game.add_moves([(0, i, 0)])
    for i in range(0, size[2], 2): game.add_moves([(2, i, 0)])
    for i in range(0, size[0], 2): game.add_moves([(0, i, 0)])
    
def play_chessboard(game):
    model = game.current_state.model
    size = model.sizes
    symmetry = [int(180 // s) for s in model.symmetry]
    
    # Moves are 3-Tuples (axis, slice, dir) axis=0,1,2 slice=0,...,size-1 dir=0,1
    if size[0] % 2 == size[1] % 2 == 1:
        moves =  [(0, i, 0) for i in range(1, size[0], 2)]
        moves += [(1, i, 0) for i in range(1, size[1], 2)]
        moves += [(2, i, 0) for i in range(1, size[2], 2)]
    elif size[1] % 2 == size[2] % 2 == 1:
        moves =  [(1, i, 0) for i in range(1, size[1], 2)]
        moves += [(2, i, 0) for i in range(1, size[2], 2)]
        moves += [(0, i, 0) for i in range(1, size[0], 2)]
    elif size[2] % 2 == size[0] % 2 == 1:
        moves =  [(2, i, 0) for i in range(1, size[2], 2)]
        moves += [(0, i, 0) for i in range(1, size[0], 2)]
        moves += [(1, i, 0) for i in range(1, size[1], 2)]
    for move in moves:
        game.add_moves([move] * symmetry[move[0]])
        
def play_t_time_Cube_odd(game):
    size = game.current_state.model.sizes
    ll = 'l{0}l{0}'.format(size[0] // 2 + 1)
    ff = 'f{0}f{0}'.format(size[0] // 2 + 1)
    game.add_flubrd("{0}du-{1}ud-{0}duu{0}uu{0}d-{0}uu".format(ll, ff))
    
def play_t_time_Cube_even(game):
    size = game.current_state.model.sizes
    ll = 'l{0}l{0}l{1}l{1}'.format(size[0] // 2, size[0] // 2 + 1)
    ff = 'f{0}f{0}f{1}f{1}'.format(size[0] // 2, size[0] // 2 + 1)
    game.add_flubrd("{0}dd2u-u2-{1}uu2d-d2-{0}dd2uuu2u2{0}uuu2u2{0}d-d2-{0}uuu2u2".format(ll, ff))
    
def play_t_time_Tower_odd(game):
    size = game.current_state.model.sizes
    ll = 'l{0}'.format(size[0] // 2 + 1)
    ff = 'f{0}'.format(size[0] // 2 + 1)
    game.add_flubrd("{0}du-{1}ud-{0}duu{0}uu{0}d-{0}uu".format(ll, ff))
    
def play_t_time_Tower_even(game):
    size = game.current_state.model.sizes
    ll = 'l{0}l{1}'.format(size[0] // 2, size[0] // 2 + 1)
    ff = 'f{0}f{1}'.format(size[0] // 2, size[0] // 2 + 1)
    game.add_flubrd("{0}dd2u-u2-{1}uu2d-d2-{0}dd2uuu2u2{0}uuu2u2{0}d-d2-{0}uuu2u2".format(ll, ff))
    
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
from ast import literal_eval
import polib # to force an import error, po_isempty ignore it if polibk is not installed

from po_isempty import po_isempty

def pprint_tranlators(translation_authors):
    print('[')
    for template, langname, langcode, persons in translation_authors:
        filename = 'po/%s.po' % langcode
        if not os.path.exists(filename):
            print('not found:', filename, file=sys.stderr)
            continue
        if po_isempty(filename, verbose=False):
            print('empty:', filename, file=sys.stderr)
            continue
        print('  ({!r}, {!r}, ['.format(langname, langcode))
        for person in persons:
            print('    {!r},'.format(person))
        print('  ]),')
    print(']')
    
def main():
    data = sys.stdin.read()
    if not data.strip():
        print('\nError: empty input, require output from "bc-lp-project-translators --py LPPROJECT"', file=sys.stderr)
        return
    try:
        tranlators = literal_eval(data)
    except (SyntaxError, ValueError):
        print('\nError: invalid syntax in input, require output from "bc-lp-project-translators --py LPPROJECT"')
        return
    try:
        pprint_tranlators(tranlators)
    except (ValueError, TypeError):
        print('\nError: malformed input, require output from "bc-lp-project-translators --py LPPROJECT"')
        raise
    
    
if __name__ == '__main__':
    main()
    




#!/usr/bin/python3
#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009, 2011-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os

import pybiklib.main

root_dir = os.path.dirname(os.path.realpath(os.path.dirname(__file__)))

def main():
    pybiklib.main.run(root_dir)

if __name__ == '__main__':
    main()




#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009, 2011-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: guile-hooks.c
# Original copyright and license: 2004  Dale Mellor, GPL3+



import os
import sys
import imp
from collections import namedtuple, OrderedDict

from .debug import *    # pylint: disable=W0401, W0614
from . import config
from .moves import MoveQueue
from . import model

FILE_VERSION = '1.1'


class PluginState (object):
    def __init__(self, app):
        self.initial_state = None
        self.current_state = app.gamestate.current_cube_state.copy()
        self.moves = MoveQueue()
        self.error_dialog = app.error_dialog
        self.set_progress = app.set_progress
        self.end_progress = app.end_progress
        
    def add_moves(self, moves):
        for move in moves:
            self.current_state.rotate_slice(move)
            self.moves.push(move)
            
    def add_flubrd(self, code):
        #TODO: here flubrd should be used, not the one in all_moves,
        # the converter in all_moves may be configurable in the future
        for res in self.moves.parse_iter(code, len(code), self.current_state.model):
            self.current_state.rotate_slice(res[0])
            
    def random(self, count=-1):
        self.current_state.set_solved()
        self.current_state.randomize(count)
        self.initial_state = self.current_state.copy()
        self.moves.reset()
        
    def test_model(self, models, show_error=True):
        cube = self.current_state
        for unused_modelstr, Model, sizes, exp in models:
            if Model == '*':
                return True
            if Model is None:
                continue
            if cube.type != Model.type:
                continue
            sizedict = {}
            for size1, size2 in zip(sizes, cube.model.sizes):
                if size1 is None:
                    continue
                if type(size1) is int:
                    if size1 != size2:
                        break
                else:
                    sizedict[size1] = size2
            else:
                try:
                    if exp is None or eval(exp, {}, sizedict):
                        return True
                except ValueError:
                    continue
        if not show_error:
            return False
        models = [model_[0] for model_ in models if model_[1] is not None]
        if not models:
            self.error_dialog(_('This algorithm does not work for any model.\n'))
        elif len(models) == 1:
            self.error_dialog(_('This algorithm only works for:\n') + '  ' + models[0])
        else:
            self.error_dialog(
                    _('This algorithm only works for:\n') +
                    '\n'.join([' • ' + m for m in models])
                )
        return False
        
        
class PluginHelper (object):
    def __init__(self):
        self.scripts = OrderedDict()
        self.error_messages = []
        
    def call(self, app, index):
        plugin_state = PluginState(app)
        all_models = []
        for models, func in list(self.scripts.values())[index]:
            if plugin_state.test_model(models, show_error=False):
                func(plugin_state)
                break
            all_models += models
        else:
            plugin_state.test_model(all_models, show_error=True)
        return plugin_state.initial_state, plugin_state.moves
        
    def load_plugins_from_directory(self, dirname):
        if not os.path.isdir(dirname):
            debug('Plugins path does not exist:', dirname)
            return
        debug('Loading Plugins from', dirname)
        sys.path.insert(0, dirname)
        for filename in sorted(os.listdir(dirname), key=str.lower):
            unused_name, ext = os.path.splitext(filename)
            if ext != '.algorithm':
                continue
            if DEBUG_ALG: debug('  algorithm:', filename)
            try:
                self.load_text_plugin(os.path.join(dirname, filename))
            except Exception as e:      # pylint: disable=W0703
                self.error_messages.append('Error loading {}:\n{}'.format(os.path.basename(filename), e))
                sys.excepthook(*sys.exc_info())
        
    def load_plugins(self):
        '''This function initializes the plugins for us, and once the plugins have
        all been imported, it returns the requested menu structure to the caller.'''
        self.scripts.clear()
        del self.error_messages[:]
        self.load_plugins_from_directory(config.SCRIPT_DIR)
        self.load_plugins_from_directory(config.USER_SCRIPT_DIR)
        if DEBUG_ALG: debug('found', len(self.scripts))
        return [(path, i) for i, path in enumerate(self.scripts.keys())]
        
    @staticmethod
    def parse_text_plugin(filename):
        with open(filename, 'rt', encoding='utf-8') as fp:
            lines = fp.readlines()
        paras = []
        
        # parse file
        para = {}
        key = None
        for line in lines:
            line = line.rstrip()
            if not line:
                # end para
                if para:
                    paras.append(para)
                para = {}
            elif line[0] == '#':
                # comment
                pass
            elif line[0] in ' \t':
                line = line.strip()
                if para and line[0] != '#':
                    # multiline
                    para[key].append(line)
            else:
                key, value = line.split(':')
                value = value.strip()
                para[key] = [value] if value else []
        if para:
            paras.append(para)
        return paras
        
    @staticmethod
    def eval_model(modelstrings):
        model_infos = []
        for modelstr in modelstrings:
            try:
                model_info = model.from_string(modelstr)
            except ValueError as e:
                debug('Error in model %s:' % modelstr, e)
            else:
                model_infos.append(model_info)
        return model_infos
        
    @classmethod
    def eval_header(cls, header):
        value = header.get('File-Version', [])
        if len(value) == 0:
            value = FILE_VERSION
            debug('No file version found, assume version', value)
        elif len(value) == 1:
            value = value[0]
        else:
            debug('Multiple version numbers found:\n ', '\n  '.join(value))
            value = value[0]
            debug('assume version', value)
        if value != FILE_VERSION:
            debug('Wrong file version:', value)
            return None, None
        models = header.get('Model', [])
        models = cls.eval_model(models)
        
        ref_blocks = header.get('Ref-Blocks', None)
        if ref_blocks:
            ref_blocks = ' '.join(ref_blocks).split()
        return models, ref_blocks
        
    @classmethod
    def eval_paras(cls, models, ref_blocks, paras):
        for para in paras:
            value = para.get('Path', ())
            if len(value) != 1 or not value[0]:
                if DEBUG_ALG: debug('    skip Path:', value)
                para.clear()
                continue
            def split_path(value):
                if not value:
                    return None, ()
                sep = value[0]
                value = value.strip(sep).split(sep)
                return sep, tuple(v for v in value if v)
            sep, value = split_path(value[0])
            if not value:
                if DEBUG_ALG: debug('    skip Path:', value)
                para.clear()
                continue
            if DEBUG_ALG: debug('    Path:', value)
            para['Path'] = sep, value
            
            value = para.get('Depends', [])
            value = [split_path(v)[1] for v in value]
            para['Depends'] = value
            
            value = para.get('Model', None)
            para['Model'] = models if value is None else cls.eval_model(value)
            
            value = para.get('Ref-Blocks', None)
            if value:
                value = ' '.join(value).split()
            para['Ref-Blocks'] = value or ref_blocks
            
            value = para.get('Solution', None)
            def split_solution(value):
                value = value.split('#', 1)[0]
                value = value.split(',', 1)
                if len(value) != 2:
                    return None
                cond = value[0].strip().split(' ')
                cond = [c.split('=') for c in cond if c]
                cond = [c for c in cond if len(c) == 2]
                return cond, value[1].strip()
            if value is not None:
                value = [split_solution(v) for v in value if v]
            para['Solution'] = value
            
            value = para.get('Moves', None) or None
            if value is not None:
                value = ' '.join(value)
            para['Moves'] = value
            
            value = para.get('Module', None) or None
            if value is not None:
                value = value[0]
            para['Module'] = value
            
    def load_text_plugin(self, filename):
        paras = self.parse_text_plugin(filename)
        if not paras:
            return
            
        # evaluate header
        models, ref_blocks = self.eval_header(paras[0])
        if models is None:
            return
            
        # evaluate solutions
        self.eval_paras(models, ref_blocks, paras)
        
        scripts = []
        for para in paras:
            if not para:
                continue
            sep, path = para['Path']
            models = para['Model']
            depends = para['Depends']
            solution = para['Solution']
            moves = para['Moves']
            module = para['Module']
            if depends or solution is not None:
                if moves is not None:
                    debug('    solution can not have moves:', path)
                if module is not None:
                    debug('    solution can not have a module field:', path)
                params = ScriptParams(
                                depends=depends,
                                ref_blocks=para['Ref-Blocks'],
                                solution=solution,
                                scripts=scripts,
                                models=models,
                                sep_path=(sep, path),
                            )
                scripts.append((path, sep!='@', models, ScriptFactory(self, params)))
            elif moves is not None:
                if module is not None:
                    debug('    Moves can not have a module field:', path)
                def play_moves(moves, models):
                    def func(game):
                        if game.test_model(models):
                            game.add_flubrd(moves)
                    return func
                scripts.append((path, True, models, play_moves(moves, models)))
            elif module is not None:
                dirname = os.path.dirname(filename)
                modulename, funcname = module.split(' ', 1)
                mfile, mpath, mdesc = imp.find_module(modulename, [dirname])
                moduleobj = imp.load_module(modulename, mfile, mpath, mdesc)
                modulefunc = getattr(moduleobj, funcname)
                scripts.append((path, True, models, modulefunc))
            else:
                debug('    skip Path without algorithm:', path)
            
        for path, visible, models, func in scripts:
            if visible:
                funclist = self.scripts.setdefault(path, [])
                funclist.append((models, func))
        
        
ScriptParams = namedtuple('ScriptParams', 'depends ref_blocks solution scripts models sep_path')

class ScriptFactory (object):
    def __init__(self, plugin, params):
        self.solved_face_colors = {}
        self.plugin = plugin
        self.params = params
        
    def __call__(self, plugin_state):
        if not plugin_state.test_model(self.params.models):
            return
            
        depends = list(reversed(self.params.depends))
        scripts = {path: func for path, visible, models, func in self.params.scripts}
        for depend in depends:
            instance = scripts[depend]
            depends += instance.params.depends
        for depend in reversed(depends):
            instance = scripts[depend]
            self.execute(plugin_state, instance.params)
        self.execute(plugin_state, self.params)
        
    def test_face(self, cube, blocksym, condition, face):
        try:
            color2 = self.solved_face_colors[condition[face]]
        except KeyError:
            return True
        blocknum = cube.model.block_symbolic_to_block_index(blocksym)
        color1 = cube.get_colorsymbol(blocknum, blocksym[face])
        return color1 == color2
        
    def test_basic_condition(self, cube, position, condition):
        assert len(position) == len(condition)
        for i in range(len(position)):
            if not self.test_face(cube, position, condition, i):
                return False
        return True
        
    @staticmethod
    def opposite(face):
        return {  'f': 'b', 'b': 'f',
                  'l': 'r', 'r': 'l',
                  'u': 'd', 'd': 'u',
               }[face]
        
    def test_pattern_condition(self, cube, position, condition):
        if '?' in condition:
            conditions = (condition.replace('?', face, 1)
                            for face in 'flubrd'
                                if face not in condition
                                if self.opposite(face) not in condition)
            return self.test_or_conditions(cube, position, conditions)
        else:
            return self.test_basic_condition(cube, position, condition)
            
    @staticmethod
    def rotated_conditions(condition):
        for i in range(len(condition)):
            yield condition[i:] + condition[:i]
        
    def test_prefix_condition(self, cube, position, condition):
        if condition.startswith('!*'):
            return not self.test_or_conditions(cube, position, self.rotated_conditions(condition[2:]))
        elif condition.startswith('*'):
            #TODO: Instead of testing only rotated conditions, test all permutations.
            #      This should not break existing rules, and would allow to match
            #      e.g. dfr and dfl. Could be done by comparing sorted strings after
            #      expanding patterns.
            return self.test_or_conditions(cube, position, self.rotated_conditions(condition[1:]))
        elif condition.startswith('!'):
            return not self.test_pattern_condition(cube, position, condition[1:])
        else:
            return self.test_pattern_condition(cube, position, condition)
        
    def test_or_conditions(self, cube, position, conditions):
        for prefix_cond in conditions:
            if self.test_prefix_condition(cube, position, prefix_cond):
                return True
        return False
        
    def test_and_conditions(self, cube, conditions):
        for position, or_cond in conditions:
            if not self.test_or_conditions(cube, position, or_cond.split('|')):
                return False
        return True
        
    def execute(self, plugin_state, params):
        rules = params.solution
        if rules is None:
            return
        cube = plugin_state.current_state
        count = 0
        pos = 0
        while pos < len(rules):
            self.solved_face_colors.clear()
            for block in params.ref_blocks:
                blocknum = cube.model.block_symbolic_to_block_index(block)
                for face in block:
                    self.solved_face_colors[face] = cube.get_colorsymbol(blocknum, face)
            conditions, moves = rules[pos]
            if self.test_and_conditions(cube, conditions):
                if DEBUG_ALG:
                    print('{}: accept: {:2}. {}, {}'.format(
                                params.sep_path[0].join(params.sep_path[1]),
                                pos+1,
                                ' '.join('='.join(c) for c in conditions),
                                moves))
                if moves == '@@solved':
                    return
                if count > 4 * len(rules): # this value is just guessed
                    plugin_state.error_dialog(
                        'An infinite loop was detected. '
                        'This is probably an error in the solution.')
                    return
                count += 1
                
                plugin_state.add_flubrd(moves)
                pos = 0
            else:
                pos += 1
        plugin_state.error_dialog(
            'No matching rules found. '
            'This is probably an error in the solution.')
        
        



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import unicodedata


symbols = [
    #('ATOM SYMBOL',         'DejaVu-Sans-Oblique',  300, '+0+3'),
    ('ATOM SYMBOL',         'FreeSerif-Mittel',     256, '+0-6'),
    #('BEAMED EIGHTH NOTES', 'DejaVu-Sans-Oblique',  256, '+0-6'),
    ('BEAMED EIGHTH NOTES', 'FreeSerif-Mittel',     256, '+0+16'),
    ('BLACK SMILING FACE',  'DejaVu-Sans-Oblique',  256, '-1+5'),
    ('PEACE SYMBOL',        'DejaVu-Sans-Oblique',  300, '-2+5'),
    ('SHAMROCK',            'DejaVu-Sans-Oblique',  300, '-1+5'),
    ('SKULL AND CROSSBONES','DejaVu-Sans-Oblique',  300, '-1+5'),
    #('SKULL AND CROSSBONES','FreeSerif-Mittel',     300, '+0+0'),
    #('SNOWFLAKE',           'DejaVu-Sans-Oblique',  300, '-1+5'),
    ('SNOWFLAKE',           'FreeSerif-Mittel',     320, '+0+14'),
    ('WHITE SMILING FACE',  'DejaVu-Sans-Oblique',  256, '-1+5'),
    #('WHITE SUN WITH RAYS', 'DejaVu-Sans-Oblique',  300, '+0+4'),
    ('WHITE SUN WITH RAYS', 'FreeMono-Standard',    390, '+0+0'),
    ('YIN YANG',            'DejaVu-Sans-Oblique',  300, '-2+5'),
    #'FOUR LEAF CLOVER',
    #'ANTICLOCKWISE GAPPED CIRCLE ARROW',
    #'CLOCKWISE GAPPED CIRCLE ARROW',
]


path = 'data/ui/images'

def main():
    for name, font, size, pos in symbols:
        c = unicodedata.lookup(name)
        os.system('convert -size 256x256 null: -alpha transparent -font {} -pointsize {} -gravity center'
                  ' -annotate 0x0{} {} "{}/{}.png"'
                  .format(font, size, pos, c, path, name)
                  .encode('utf-8'))
    
    
if __name__ == '__main__':
    main()
    




#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009, 2011-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
from ast import literal_eval
import errno

from PyQt4.QtCore import * # pylint: disable=W0614,W0401
from PyQt4.QtCore import pyqtSignal as Signal

from .debug import debug
from . import config
from . import model


N_ = lambda s: s

def tuple_validator(lenght, itemtype, value, valuerange=None):
    if type(value) is not tuple:
        return False
    if len(value) != lenght:
        return False
    for v in value:
        if type(v) is not itemtype:
            return False
        if valuerange is not None and not valuerange[0] <= v <= valuerange[1]:
            return False
    return True
    
class KeyStore (QObject):
    schema = {
        # key:              (default,   range/enum/validator)
        #                               None: value without restriction
        #                               tuple: contains two values (min, max)
        #                               list: contains strings for the enum text,
        #                                     the index is the enum value
        #                               function: returns True, if value is valid
        'window.size':          ((480, 360),lambda v: tuple_validator(2, int, v)),
        'window.divider':       (300,       lambda v: type(v) is int),
        'window.toolbar':       (True,      lambda v: type(v) is bool),
        'window.editbar':       (True,      lambda v: type(v) is bool),
        'window.statusbar':     (True,      lambda v: type(v) is bool),
        'draw.default_rotation':((-30.,39.),lambda v: tuple_validator(2, float, v)),
        'draw.lighting':        (True,     lambda v: type(v) is bool),
        'draw.selection':       (1,         ['quadrant', 'simple']),
        'draw.speed':           (30,        (1, 100)),
        'game.type':            (0,         [m.type for m in model.models]),
        'game.size':            ((3,4,2),   lambda v: tuple_validator(3, int, v, (1, 10))),
        'game.blocks':          ('solved',  lambda v: type(v) is str),
        'game.moves':           ('',        lambda v: type(v) is str),
        'game.position':        (0,         lambda v: type(v) is int and v >= 0),
        'theme.face.0.color':   ('#a81407', lambda v: type(v) is str),
        'theme.face.1.color':   ('#d94b1c', lambda v: type(v) is str),
        'theme.face.2.color':   ('#f0c829', lambda v: type(v) is str),
        'theme.face.3.color':   ('#e3e3e3', lambda v: type(v) is str),
        'theme.face.4.color':   ('#1d6311', lambda v: type(v) is str),
        'theme.face.5.color':   ('#00275e', lambda v: type(v) is str),
        'theme.face.#.image':   ('',        lambda v: type(v) is str),
        'theme.face.#.mode':    (0,         ['tiled', 'mosaic']),
        'theme.bgcolor':        ('#B9a177', lambda v: type(v) is str),
        'draw.accels':          ([('r', 'KP+6'), ('r-', 'Shift+KP+Right'),
                                  ('l', 'KP+4'), ('l-', 'Shift+KP+Left'),
                                  ('u', 'KP+8'), ('u-', 'Shift+KP+Up'),
                                  ('d', 'KP+2'), ('d-', 'Shift+KP+Down'),
                                  ('f', 'KP+5'), ('f-', 'Shift+KP+Clear'),
                                  ('b', 'KP+0'), ('b-', 'Shift+KP+Ins'),
                                  ('R', 'Ctrl+KP+8'), ('L', 'Ctrl+KP+2'),
                                  ('U', 'Ctrl+KP+4'), ('D', 'Ctrl+KP+6'),
                                  ('F', 'Ctrl+KP+5'), ('B', 'Ctrl+KP+0'),
                                 ],         lambda v: type(v) is list),
        'draw.zoom':            (1.0,       (0.1, 100.0)),
        # The following 6 words are for the antialiasing levels: disabled, ugly, low, medium, high, higher
        'draw.samples':         (3,         [N_('disabled'), N_('ugly'), N_('low'),
                                             N_('medium'), N_('high'), N_('higher')]),
        'draw.mirror_faces':    (False,     lambda v: type(v) is bool),
        'draw.mirror_distance': (2.1,       (0.1, 10.0)),
        'action.edit_moves':    ('Ctrl+L',  lambda v: type(v) is str),
        'action.edit_model':    ('',        lambda v: type(v) is str),
        'action.reload_shader': ('',        lambda v: type(v) is str),
        'action.selectmodel':   ('',        lambda v: type(v) is str),
        'action.initial_state': ('',        lambda v: type(v) is str),
        'action.reset_rotation':('',        lambda v: type(v) is str),
        'action.invert_moves':  ('',        lambda v: type(v) is str),
        'action.reload_scripts':('',        lambda v: type(v) is str),
        'action.preferences':   ('',        lambda v: type(v) is str),
        'action.normalize_complete_rotations':('',lambda v: type(v) is str),
       }
       
    changed = Signal(str)
    error = Signal(str)
    
    def __init__(self, filename):
        QObject.__init__(self)
        self.filename = filename
        self.keystore = {}
        self.read_settings()
        self.write_timer = QTimer(self)
        self.write_timer.setSingleShot(True)
        self.write_timer.setInterval(5000)
        self.write_timer.timeout.connect(self.write_settings)
        
    def get_default(self, key):
        return self.schema[key][0]
    def get_range(self, key):
        return self.schema[key][1]
    
    def get_value(self, key):
        try:
            return self.keystore[key]
        except KeyError:
            return self.get_default(key)
            
    def get_nick(self, key):
        value = self.get_value(key)
        valuerange = self.get_range(key)
        if not isinstance(valuerange, list):
            raise ValueError('{} is not an enum'.format(key))
        return valuerange[value]
        
    def set_value(self, key, value):
        self.keystore[key] = value
        self.changed.emit(key)
        if not self.write_timer.isActive():
            self.write_timer.start()
        
    def set_nick(self, key, nick):
        valuerange = self.get_range(key)
        if not isinstance(valuerange, list):
            raise ValueError('{} is not an enum'.format(key))
        value = valuerange.index(nick)
        return self.set_value(key, value)
        
    def del_value(self, key):
        try:
            del self.keystore[key]
        except KeyError:
            pass # already the default value
        self.changed.emit(key)
        if not self.write_timer.isActive():
            self.write_timer.start()
        
    def read_settings(self):
        if not self.filename:
            return
        keys = list(self.schema.keys())
        dirname = os.path.dirname(self.filename)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
            
        # Only convert from gconf to standard location
        write_version_file = False
        if self.filename == config.USER_SETTINGS_FILE and not os.path.exists(config.USER_VERSION_FILE):
            write_version_file = True
            try:
                if not os.path.exists(self.filename):
                    from . import migration
                    migration.gconf_0_5_to_settings_1_0(self.filename)
            except Exception:   # pylint: disable=W0703
                # Never fail, but report error
                sys.excepthook(*sys.exc_info())
                
        # read settings
        try:
            with open(self.filename, 'rt', encoding='utf-8') as settings_file:
                lines = settings_file.readlines()
        except IOError as e:
            if e.errno == errno.ENOENT:
                lines = []
            else:
                raise
        for line in lines:
            # parse the line, discard invalid keys
            try:
                key, strvalue = line.split('=', 1)
            except ValueError:
                continue
            key = key.strip()
            strvalue = strvalue.strip()
            if key not in keys:
                continue
            try:
                value = literal_eval(strvalue)
            except (ValueError, SyntaxError):
                continue
                
            # translate enums and validate values
            valuerange = self.get_range(key)
            if isinstance(valuerange, list):
                try:
                    value = valuerange.index(value)
                except ValueError:
                    continue
            elif isinstance(valuerange, tuple):
                if not valuerange[0] <= value <= valuerange[1]:
                    continue
            elif valuerange is not None:
                if not valuerange(value):
                    continue
                    
            self.keystore[key] = value
            
        if write_version_file:
            with open(config.USER_VERSION_FILE, 'wt', encoding='utf-8') as version_file:
                version_file.write(config.VERSION)
                
    def dump(self, file, all=False):    # pylint: disable=W0622
        keydict = self.schema if all else self.keystore
        for key in sorted(keydict.keys()):
            if '#' in key:
                continue
            value = self.get_value(key)
            # translate enums
            valuerange = self.get_range(key)
            if isinstance(valuerange, list):
                value = valuerange[value]
            print(key, '=', repr(value), file=file)
        
    def write_settings(self):
        if not self.filename:
            return
        try:
            with open(self.filename, 'wt', encoding='utf-8') as settings_file:
                self.dump(settings_file)
        except EnvironmentError as e:
            error_message = _('Settings can not be written to file: '
                            '{error_message}').format(error_message=e)
            debug(error_message)
            self.error.emit(error_message)
            
        
class Settings (object):
    keystore = None
    
    def __init__(self, key=''):
        object.__setattr__(self, '_key', key)
        object.__setattr__(self, '_array_len', 0)
        
    def load(self, filename):
        self.__class__.keystore = KeyStore(filename)
        keys = list(self.keystore.schema.keys())
        deferred = []
        
        for key in keys:
            subkeys = key.split('.')
            settings_parent = self
            for subkey in subkeys[:-1]:
                if subkey == '#':
                    deferred.append(key)
                    break
                if subkey.isdigit():
                    subkey = str(int(subkey))
                try:
                    settings_child = getattr(settings_parent, subkey)
                except KeyError:
                    settings_child = Settings(settings_parent._key + subkey + '.')
                    object.__setattr__(settings_parent, subkey, settings_child)
                assert isinstance(settings_child, Settings)
                settings_parent = settings_child
            else:
                object.__setattr__(settings_parent, subkeys[-1]+'_range', self.keystore.get_range(key))
        changed = False
        for key in deferred:
            subkeys = key.split('.')
            settings_parent = self
            for i, subkey in enumerate(subkeys):
                if subkey == '#':
                    for subkey in list(settings_parent.__dict__.keys()):
                        if subkey[0] == '_':
                            continue
                        expanded = '.'.join(subkeys[:i] + [subkey] + subkeys[i+1:])
                        if expanded not in self.keystore.schema:
                            changed = True
                            self.keystore.schema[expanded] = self.keystore.schema[key]
                    break
                settings_parent = getattr(settings_parent, subkey)
        if changed:
            self.load(filename)
            
    def close(self):
        self.keystore.write_timer.stop()
        self.keystore.write_settings()
        
    def __getattr__(self, key):
        if key.endswith('_nick'):
            return self.keystore.get_nick(self._key + key[:-5])
        return self.keystore.get_value(self._key + key)
        
    def __getitem__(self, key):
        return getattr(self, str(key))
        
    def __setattr__(self, key, value):
        if key.endswith('_nick'):
            key = self._key + key[:-5]
            func = self.keystore.set_nick
        else:
            key = self._key + key
            func = self.keystore.set_value
        if key in list(self.keystore.schema.keys()):
            func(key, value)
        else:
            raise AttributeError('use object.__setattr__ to set attributes')
            
    def __delattr__(self, key):
        _key = self._key + key
        if _key in list(self.keystore.schema.keys()):
            self.keystore.del_value(_key)
        else:
            raise AttributeError('use object.__delattr__ to delete attributes')
            

settings = Settings()




# -*- coding: utf-8 -*-

#  Copyright © 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


cdef extern from 'glm/glm.hpp' namespace "glm":
    cdef cppclass vec3:
        float x, y, z
        vec3()
        vec3(float, float, float)
        
    cdef cppclass vec4:
        float x, y, z, w
        vec4()
        vec4(float, float, float, float)
        float& operator[](int)
        vec4 operator+(vec4)
        vec4 operator*(vec4)
        
    cdef cppclass mat4:
        mat4()
        mat4(float)
        vec4& operator[](int)
        mat4 operator*(mat4)
        vec4 operator*(vec4)
        
cdef extern from 'glm/gtx/transform.hpp' namespace "glm":
    cdef mat4 rotate(float, vec3)
    
    




#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: move-queue.c
# Original copyright and license: 2004  Dale Mellor, GPL3+



import re
from collections import namedtuple

from .debug import *    # pylint: disable=W0401, W0614


# A structure containing information about a movement taking place.
class MoveData (namedtuple('_MoveData', 'axis slice dir')):     # pylint: disable=W0232
    # axis  = 0...
    # slice = 0...dim-1 or -1 for all slices
    # dir   = 0 or 1
    def inverted(self):
        return self._replace(dir=not self.dir)  # pylint: disable=E1101

class MoveQueue(object):
    class MoveQueueItem(object):
        def __init__(self, data, mark_after=False, code=None, markup=None):
            self.data = MoveData(*data) or MoveData(0, 0, 0)
            self.mark_after = mark_after
            self.code = code
            self.markup = markup
        axis = property(lambda self: self.data.axis)
        slice = property(lambda self: self.data.slice)
        dir = property(lambda self: self.data.dir)
        def copy(self):
            return MoveQueue.MoveQueueItem(self.data,
                    mark_after=self.mark_after,
                    code=self.code,
                    markup=self.markup)
        def invert(self):
            self.data = self.data.inverted()
        def rotate_by(self, model, totalmove):
            rmove = model.rotate_move(totalmove.data, self.data)
            self.data = MoveData(*rmove)
            self.code = None
            self.markup = None
            
    def __init__(self):
        self.current_place = 0      # Number of steps of current from head.
        self.moves = []
        self.converter = FormatFlubrd
    def copy(self):
        movequeue = MoveQueue()
        movequeue.current_place = self.current_place
        movequeue.moves = [m.copy() for m in self.moves]
        movequeue.converter = self.converter
        return movequeue
        
    def at_start(self):
        return self.current_place == 0
    def at_end(self):
        return self.current_place == self.queue_length
    @property
    def _prev(self):
        return self.moves[self.current_place-1]
    @property
    def _current(self):
        return self.moves[self.current_place]
    @property
    def queue_length(self):
        return len(self.moves)
    def __str__(self):
        return 'MoveQueue(len=%s)' % self.queue_length

    # Routine to push a new MoveData object onto the _back_ of the given queue. A
    # new,  locally-managed copy is made of the incumbent datum. Return 1 (one) if
    # the operation is successful; 0 (zero) indicates that there was insufficient
    # memory to grow the queue.
    def push(self, move_data, mark_after=False, code=None, markup=None):
        new_element = self.MoveQueueItem(move_data,
                                mark_after=mark_after, code=code, markup=markup)
        self.moves.append(new_element)

    # Procedure to copy the incumbent datum to the current place in the queue,  and
    # drop all subsequent queue entries (so that the current element becomes the
    # tail). If current points past the end of the tail (including the case when
    # there are no data in the queue),  then a regular push operation is performed
    # at the tail. In all other cases 1 (one) will be returned.
    def push_current(self, move_data):
        # If there are no data in the queue,  then perform a standard push
        # operation. Also do this if we are at the tail.
        if not self.at_end():
            self.moves[self.current_place:] = []
        self.push(move_data)

    # Routine to get the data at the 'current' position in the queue. If there are
    # no data in the queue,  None will be returned.
    def current(self):
        return None if self.at_end() else self._current.data

    # Routine to retard the current pointer (move it towards the _head_ of the queue).
    def retard(self):
        if not self.at_start():
            self.current_place -= 1
    
    def rewind_start(self):
        self.current_place = 0
        
    def forward_end(self):
        self.current_place = self.queue_length
        
    # Remove the current object and all those that come afterwards.
    def truncate(self):
        self.moves[self.current_place:] = []
        
    def truncate_before(self):
        self.moves[:self.current_place] = []
        self.current_place = 0
        
    def reset(self):
        self.current_place = 0
        self.moves[:] = []
        
    def advance(self):
        if not self.at_end():
            self.current_place += 1
        
    def swapnext(self):
        cp = self.current_place
        if cp+1 < self.queue_length:
            ma, mb = self.moves[cp], self.moves[cp+1]
            ma.code = mb.code = None
            ma.data, mb.data = mb.data, ma.data
            self.advance()
            
    def swapprev(self):
        cp = self.current_place
        if 0 < cp < self.queue_length:
            ma, mb = self.moves[cp-1], self.moves[cp]
            ma.code = mb.code = None
            ma.data, mb.data = mb.data, ma.data
            self.retard()
            
    def invert(self):
        # a mark at the end of the moves is discarded because a mark at start is not supported
        mark = False
        for move in self.moves:
            move.invert()
            move.code = None
            move.markup = None
            move.mark_after, mark = mark, move.mark_after
        self.moves.reverse()
        if not (self.at_start() or self.at_end()):
            self.current_place = len(self.moves) - self.current_place
            
    def normalize_complete_rotations(self, model):
        totalmoves = []
        new_moves = []
        for i, move in enumerate(self.moves):
            if i == self.current_place:
                self.current_place = len(new_moves)
            if move.slice < 0:
                totalmoves.append(move)
            else:
                for totalmove in reversed(totalmoves):
                    move.rotate_by(model, totalmove)
                new_moves.append(move)
        self.moves = new_moves + totalmoves
        
    def is_mark_current(self):
        return self.at_start() or self._prev.mark_after
        
    def is_mark_after(self, pos):
        return self.moves[pos].mark_after
        
    def mark_current(self, mark=True):
        if not self.at_start():
            self._prev.mark_after = mark
            if self._prev.code is not None:
                self._prev.code = self._prev.code.replace(' ','')
                #FIXME: recreate self._prev.markup
                if mark:
                    self._prev.code += ' '
                    
    def mark_and_extend(self, other):
        if not other.moves:
            return -1
        self.mark_current()
        self.truncate()
        self.moves += other.moves
        return self.current_place + other.current_place
        
    def format(self, model):
        #TODO: arg to use explicit converter
        code = ''
        markup = ''
        pos = 0
        for i, move in enumerate(self.moves):
            if move.code is None:
                move.code, move.markup = self.converter.format(move, model)
            code += move.code
            markup += move.markup
            if i < self.current_place:
                pos = len(code)
        return code, pos, markup
        
    def parse_iter(self, code, pos, model):
        #TODO: arg to use explicit converter
        code = code.lstrip(' ')
        queue_pos = self.current_place
        move_code = ''
        for i, c in enumerate(code):
            if move_code and self.converter.isstart(c, model):
                data, mark, markup_code = self.converter.parse_move(move_code, model)
                if data is not None:
                    #FIXME: invalid chars at start get lost, other invalid chars are just ignored
                    self.push(data, mark, move_code, markup_code)
                yield data, queue_pos, i
                if i == pos:
                    queue_pos = self.queue_length
                move_code = c
            else:
                move_code += c
            if i < pos:
                queue_pos = self.queue_length + 1
        if move_code:
            data, mark, markup_code = self.converter.parse_move(move_code, model)
            if data is not None:
                self.push(data, mark, move_code, markup_code)
            if len(code)-len(move_code) < pos:
                queue_pos = self.queue_length
            yield data, queue_pos, len(code)
            
    def parse(self, code, pos, model):
        qpos = 0
        cpos = 0
        for res in self.parse_iter(code, pos, model):
            qpos = res[1]
            if cpos < pos:
                cpos = res[2]
        return qpos, cpos
        
        
class FormatFlubrd: # pylint: disable=W0232
    re_flubrd = re.compile(r"(.)(\d*)(['-]?)([^ ]*)( *)(.*)")
    
    @classmethod
    def isstart(cls, char, model):
        return char.upper() in model.faces
    
    @staticmethod
    def _format_markup(mface, mslice, mdir, err1, mark, err2):
        if err1:
            err1 = '<span underline="error" color="red">%s</span>' % err1
        if err2:
            err2 = '<span underline="error">%s</span>' % err2
        return ''.join((mface, mslice, mdir, err1, mark, err2))
    
    @staticmethod
    def intern_to_normal_move(maxis, mslice, mdir, model):
        if mslice == -1:
            return model.symbolsI[maxis] if mdir else model.symbols[maxis], 0, 0, True
        elif mslice*2 > model.sizes[maxis]-1:
            return model.symbolsI[maxis], model.sizes[maxis]-1 - mslice, 1-mdir, False
        else:
            return model.symbols[maxis], mslice, mdir, False
    
    @classmethod
    def format(cls, move, model):
        mface, mslice, mdir, whole_cube = cls.intern_to_normal_move(move.axis, move.slice, move.dir, model)
        mface = mface.upper() if whole_cube else mface.lower()
        mslice = str(mslice+1) if mslice else ''
        mdir = '-' if mdir else ""
        mark = ' ' if move.mark_after else ''
        move_code = mface + mslice + mdir + mark
        markup_code = cls._format_markup(mface, mslice, mdir, '', mark, '')
        return move_code, markup_code
        
    @staticmethod
    def normal_to_intern_move(mface, mslice, mdir, whole_cube, model):
        if mface in model.symbols:
            return model.symbols.index(mface), (-1 if whole_cube else mslice), mdir
        elif mface in model.symbolsI:
            maxis = model.symbolsI.index(mface)
            return maxis, (-1 if whole_cube else model.sizes[maxis]-1 - mslice), 1-mdir
        else:
            return None, None, None
            
    @classmethod
    def parse_move(cls, move_code, model):
        #debug('move_code: '+move_code)
        mface, mslice, mdir, err1, mark, err2 = cls.re_flubrd.match(move_code).groups()
        markup_code = cls._format_markup(mface, mslice, mdir, err1, mark, err2)
        whole_cube = mface.isupper()
        mface = mface.upper()
        mslice = int(mslice or 1) - 1
        mdir = int(bool(mdir))
        mark = bool(mark)
        # move finished, normalize it
        maxis, mslice2, mdir = cls.normal_to_intern_move(mface, mslice, mdir, whole_cube, model)
        if maxis is not None and 0 <= mslice < model.sizes[maxis]:
            return MoveData(maxis, mslice2, mdir), mark, markup_code
        debug('Error parsing formula')
        return None, False, move_code
        
    @staticmethod
    def parse(code, pos, model):
        moves = MoveQueue()
        qpos, cpos = moves.parse(code, pos, model)
        return moves, qpos, cpos
        
        
